/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * contrib/proxymon_info.c: send proxy monitor notices on connect
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: proxymon_info.c,v 1.1 2004/04/03 19:54:44 rageircd Exp $
 */

/* This module creates two new conf variables within the general {}
 * conf block. These are proxy_monitor_host and proxy_monitor_url.
 *
 * The former variable is the host from which any proxy detection
 * scans are performed. The latter variable is a URL (or point of
 * contact) for clients, regarding the proxy scans. Both variables
 * must be configured in order for such notices to be sent.
 *
 * general {
 *	proxy_monitor_host = "proxy.checker.host";
 *	proxy_monitor_url = "http://www.your-irc-network.com/proxy.html";
 * };
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include "modules.h"
#include "hook.h"
#include "conf2.h"
#include "fd.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

static int parse_pm(HookData *);
static int test_pm(HookData *);
static int rehash_pm(HookData *);
static int send_proxy_notices(HookData *);

static void free_proxymon();

static char *proxy_monitor_host = NULL;
static char *proxy_monitor_url = NULL;

Module MOD_HEADER(proxymon_info) = {
	"modules/proxymon_info.so",
	"Proxy monitor notices",
	5, "$Revision: 1.1 $"
};

int MOD_LOAD(proxymon_info)()
{
	register_hook_event(&MOD_HEADER(proxymon_info), h_conf_parse, parse_pm);
	register_hook_event(&MOD_HEADER(proxymon_info), h_conf_test, test_pm);
	register_hook_event(&MOD_HEADER(proxymon_info), h_conf_rehash, rehash_pm);
	register_hook_event(&MOD_HEADER(proxymon_info), h_post_motd, send_proxy_notices);
	return MOD_SUCCESS;
}

int MOD_UNLOAD(aup_notice)()
{
	free_proxymon();
	return MOD_SUCCESS;
}

static void free_proxymon()
{
	cMyFree(proxy_monitor_host);
	cMyFree(proxy_monitor_url);
}

static int parse_pm(HookData *hdata)
{
	ConfigUnknown *unknown = (ConfigUnknown *)hdata->v;
	ConfigEntry *entry = unknown->entry;

	if (unknown->type == CONF_GENERAL) {
		if (!mycmp(entry->varname, "proxy_monitor_host")) {
			cDupString(proxy_monitor_host, entry->vardata);
			return CONF_SUCCESS;
		}
		if (!mycmp(entry->varname, "proxy_monitor_url")) {
			cDupString(proxy_monitor_url, entry->vardata);
			return CONF_SUCCESS;
		}
	}
	return CONF_NOMATCH;
}

static int test_pm(HookData *hdata)
{
	ConfigUnknown *unknown = (ConfigUnknown *)hdata->v;

	if (unknown->type == CONF_GENERAL) {
		if (BadPtr(proxy_monitor_host) || BadPtr(proxy_monitor_url)) {
			free_proxymon();
		}
	}
	return CONF_NOMATCH;
}

static int rehash_pm(HookData *hdata)
{
	free_proxymon();
	return CONF_SUCCESS;
}

static int send_proxy_notices(HookData *hdata)
{
	aClient *sptr = (aClient *)hdata->sptr;

	if (BadPtr(proxy_monitor_host) || BadPtr(proxy_monitor_url)) {
		return 0;
	}

	send_me_noticeNA(sptr, ":*** This server runs an open proxy monitor to prevent abuse.");
	send_me_notice(sptr, ":*** If you see connections on various ports from %s", proxy_monitor_host);
	send_me_noticeNA(sptr, ":*** please disregard them, as they are the monitor in action.");
	send_me_notice(sptr, ":*** For more information, please visit/contact %s", proxy_monitor_url);

	return 0;
}
