/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_kline.c: KLINE command module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_kline.c,v 1.38 2004/05/08 16:12:02 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include "modules.h"
#include "xmode.h"
#include <utmp.h>
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

void expire_klines();

Module MOD_HEADER(m_kline) = {
	"modules/commands/m_kline.so",
	"/KLINE command",
	6, "$Revision: 1.38 $"
};

int MOD_LOAD(m_kline)()
{
	register_command(&MOD_HEADER(m_kline), &CMD_KLINE, m_kline);
	register_event(&MOD_HEADER(m_kline), "expire_klines", expire_klines, NULL, 5, 1);
	return MOD_SUCCESS;
}

int MOD_UNLOAD(m_kline)()
{
	if (Internal.next_expire_klines != -1) {
		sendto_realops("Cannot unload module %s while timed K:lines remain",
			MOD_HEADER(m_kline).name);
		return MOD_FAILURE;
	}
	return MOD_SUCCESS;
}

static int write_kline_to_file(ConfigItem_ban *ban, char *client_path)
{
	char buf[1024];
	int fd = -1, retval = 0;

	if ((fd = open(Internal.conf_file, O_RDWR|O_APPEND)) != -1) {
		ircsprintf(buf, "\n/* Permanent K:line added by %s */\nban userhost {\n\tmask = "
			"\"%s\";\n\treason = \"%s\";\n};\n", client_path, ban->mask,
			BadPtr(ban->reason) ? "no reason" : ban->reason);
		retval = write(fd, buf, strlen(buf));
		close(fd);
	}
	return retval;
}

static time_t expire_is_valid(char *expire_str)
{
	time_t expire = -1;
	if (!(expire = atime(expire_str)) && (*expire_str != '0' ||
	  (*expire_str == '0' && *(expire_str + 1) != '\0'))) {
		return -1;
	}
	return atime(expire_str);
}

/*
 * m_kline
 *	parv[0] = sender prefix
 *	parv[1] = user@host, host or nickname
 *
 *	Temporary K:line
 *		parv[2] = reason (optional)
 *	Timed K:line
 *		parv[2] = expire time (>0)
 *		parv[3] = reason (optional)
 *	Permanent K:line
 *		parv[2] = '0'
 *		parv[3] = reason (optional)
 */
int m_kline(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	char *user, *host, *userhost, *p, tmp[USERLEN + 1];
	char *reason = "No reason";
	dlink_node *node;
	ConfigItem_ban *ban;
	time_t expire = -1;
	aClient *acptr;
	int i = 0;

	if (!OPHasFlag(sptr, OFLAG_KLINE)) {
		send_me_numericNA(sptr, ERR_NOPRIVILEGES);
		return 0;
	}
	if (parc < 2 || BadPtr(parv[1])) {
		send_me_numeric(sptr, ERR_NEEDMOREPARAMS, "KLINE");
		return 0;
	}
	if ((host = strchr(parv[1], '@')) != NULL || (*parv[1] == '*' || strchr(parv[1], '.') != NULL)) {
		if (host != NULL) {
			user = parv[1];
			*host++ = '\0';
		}
		else {
			user = "*";
			host = parv[1];
		}
		if (strlen(user) > USERLEN) {
			user[USERLEN] = '\0';
		}
		if (strlen(host) > HOSTLEN) {
			host[HOSTLEN] = '\0';
		}
		if (!irccmp(user, "*")) {
			for (p = host; *p != '\0'; p++) {
				if (*p != '*' && *p != '.') {
					i++;
				}
			}
			if (i < 1) {
				send_me_noticeNA(sptr, ":What a sweeping K:line! If only your admin knew "
					"you tried that...");
				sendto_realops_lev(SPY_LEV, "%s (%s@%s) attempted to K:line %s@%s",
					sptr->name, sptr->username, MaskedHost(sptr), user, host);
				return 0;
			}
		}
	}
	else {
		if ((acptr = find_chasing(sptr, parv[1], NULL, MSG_KLINE)) == NULL) {
			return 0;
		}
 		if (acptr->user == NULL || !IsClient(acptr)) {
			return 0;
		}

		*tmp = '*';
		strcpy((tmp + 1), (*acptr->username == '~') ? (acptr->username + 1) : acptr->username);
		user = tmp;
		host = acptr->host;
	}

	userhost = make_user_host(user, host);
	DLINK_FOREACH_DATA(conf_ban_list.head, node, ban, ConfigItem_ban) {
		if (!(ban->flags & BAN_USERHOST) || (ban->flags & BAN_SVS)) {
			continue;
		}
		if (!match(ban->mask, userhost)) {
			send_me_notice(sptr, ":%s is already on the K:line list as %s (%s)", userhost,
				ban->mask, ban->item.temp ? "temporary" : "permanent");
				return 0;
		}
	}
	if (parc > 3) {
		if ((expire = expire_is_valid(parv[2])) < 0) {
			send_me_notice(sptr, ":%s is not a valid expire-time.", parv[2]);
			return 0;
		}
		reason = parv[3];
	}
	else if (parc > 2) {
		if ((expire = expire_is_valid(parv[2])) < 0) {
			reason = parv[2];
		}
	}
	if (strlen(reason) > TOPICLEN) {
		reason[TOPICLEN] = '\0';
	}

	ban = (ConfigItem_ban *)MyMallocX(sizeof(ConfigItem_ban));
	ban->flags |= BAN_USERHOST;
	DupString(ban->mask, userhost);
	if (!BadPtr(reason)) {
		DupString(ban->reason, reason);
	}

	if (expire == -1) {
		ban->item.temp = ban->kill_on_rehash = 1;

		ircdlog(LOG_BAN, "temp K:line on %s (%s) added by %s", ban->mask,
			BadPtr(ban->reason) ? "no reason" : ban->reason, get_client_name(cptr, TRUE));
		sendto_realops("%s has added a temporary K:line on %s (%s).", cptr->name,
			ban->mask, BadPtr(ban->reason) ? "no reason specified" : ban->reason);
	}
	else if (!expire) {
		char *client_path = get_client_name(cptr, SHOW_IP);

		if (GeneralConfig.write_perm_klines_to_file) {
			if (!write_kline_to_file(ban, client_path)) {
				send_me_notice(sptr, ":Couldn't append permanent K:line to %s (%s), "
					"adding with expire-time of %d seconds instead.", Internal.conf_file,
					strerror(errno), DEFAULT_KLINE_EXPIRE_TIME);
				expire = DEFAULT_KLINE_EXPIRE_TIME;
			}
			else {
				ban->in_file = ban->kill_on_rehash = 1;
				ircdlog(LOG_BAN, "perm K:line on %s (%s) added by %s written "
					"to %s", ban->mask,
					BadPtr(ban->reason) ? "no reason" : ban->reason,
					client_path, Internal.conf_file);
				sendto_realops("%s has added a permanent K:line on %s (%s), "
					"which has been written to %s.", cptr->name,
					ban->mask,
					BadPtr(ban->reason) ? "no reason specified" : ban->reason,
					Internal.conf_file);
			}
		}
		else {
			ircdlog(LOG_BAN, "perm K:line on %s (%s) added by %s", ban->mask,
				BadPtr(ban->reason) ? "no reason": ban->reason, client_path);
			sendto_realops("%s has added a permanent K:line on %s (%s).",
				cptr->name, ban->mask,
				BadPtr(ban->reason) ? "no reason specified" : ban->reason);
		}
	}
	if ((ban->expire = expire) > 0) {
		ban->set_at = timeofday;
		ircdlog(LOG_BAN, "timed K:line on %s (%s) added by %s (%ld seconds)", ban->mask,
			BadPtr(ban->reason) ? "no reason" : ban->reason, get_client_name(cptr, TRUE),
			ban->expire);
		sendto_realops("%s has added a timed K:line on %s (%s), which will expire in %s.",
			cptr->name, ban->mask, BadPtr(ban->reason) ? "no reason specified" : ban->reason,
			time_to_str(ban->expire));

		/* FIXME: move to events */
		if (Internal.next_expire_klines > (ban->set_at + ban->expire) ||
		  Internal.next_expire_klines == -1) {
			Internal.next_expire_klines = (ban->set_at + ban->expire);
		}
	}

	dlink_add(&conf_ban_list, ban);
	Internal.do_ban_check = 1;
	return 0;
}

void expire_klines()
{
	dlink_node *node, *next = NULL;
	ConfigItem_ban *ban;
	time_t new_expire_klines = -1, expire_at = 0;

	if ((Internal.next_expire_klines > timeofday) || Internal.next_expire_klines == -1) {
		return;
	}
	DLINK_FOREACH_SAFE_DATA(conf_ban_list.head, node, next, ban, ConfigItem_ban) {
		if (!ban->expire) {
			continue;
		}

		expire_at = (ban->set_at + ban->expire);
		if (Internal.next_expire_klines >= expire_at) {
			sendto_realops("Timed K:line on %s (set %s ago) has now expired.",
				ban->mask, time_to_str(timeofday - ban->set_at));
			dlink_del(&conf_ban_list, NULL, node);
			MyFree(ban->mask);
			if (!BadPtr(ban->reason)) {
				MyFree(ban->reason);
			}
			MyFree(ban);
		}
		else if (new_expire_klines > expire_at || new_expire_klines == -1) {
			new_expire_klines = expire_at;
		}
	}
	Internal.next_expire_klines = new_expire_klines;
	return;
}
