/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_message.c: PRIVMSG and NOTICE command module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_message.c,v 1.70 2004/05/13 22:11:45 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include "hook.h"
#include "modules.h"
#include "xmode.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

Hook *h_privmsg = NULL;
Hook *h_dccsend = NULL;

Module MOD_HEADER(m_message) = {
	"modules/commands/m_message.so",
	"/PRIVMSG and /NOTICE commands",
	6, "$Revision: 1.70 $"
};

int MOD_LOAD(m_message)()
{
	h_privmsg = register_hook(&MOD_HEADER(m_message), "h_privmsg");
	h_dccsend = register_hook(&MOD_HEADER(m_message), "h_dccsend");

	register_command(&MOD_HEADER(m_message), &CMD_PRIVMSG, m_privmsg);
	register_command(&MOD_HEADER(m_message), &CMD_PRIVMSG_D, m_privmsg_directed);
	register_command(&MOD_HEADER(m_message), &CMD_NOTICE, m_notice);
	register_command(&MOD_HEADER(m_message), &CMD_NOTICE_D, m_notice_directed);

	MOD_SET_FLAG(&MOD_HEADER(m_message), MOD_FLAG_PERM);
	return MOD_SUCCESS;
}

int MOD_UNLOAD(m_message)()
{
	return MOD_SUCCESS;
}

enum {
	CTCP_NONE,
	CTCP_YES,
	CTCP_NO,
	CTCP_DCC,
	CTCP_DCCSEND
};

/* FIXME -- this needs some serious reworking */

static inline int check_for_client_flood(int notice, aClient *sptr, aClient *acptr)
{
	int delta = 0;

	if (!FloodConfig.msg_flood_count || sptr->user == NULL || acptr->user == NULL
	  || IsULine(acptr) || IsServer(acptr)) {
		return 0;
	}
	if (acptr->user->first_message_time + 1 < timeofday) {
		delta = timeofday - acptr->user->first_message_time;
		acptr->user->number_of_messages -= delta;
		acptr->user->first_message_time = timeofday;

		if (acptr->user->number_of_messages <= 0) {
			acptr->user->number_of_messages = 0;
			acptr->user->flood_notice = 0;
		}
	}
	if (acptr->user->number_of_messages >= FloodConfig.msg_flood_count || acptr->user->flood_notice) {
		if (!acptr->user->flood_notice) {
			sendto_realops_lev(FLOOD_LEV, "Possible flooder: %s (%s@%s) target: %s (%s@%s)",
				sptr->name, sptr->username, MaskedHost(sptr),
				acptr->name, acptr->username, MaskedHost(acptr));
			acptr->user->flood_notice = 1;
			acptr->user->number_of_messages += 2;
		}
		if (MyClient(sptr) && !notice) {
			send_me_notice(sptr, ":Messages to %s throttled due to flooding",
				acptr->name);
		}
		return 1;
	}
	acptr->user->number_of_messages++;
	return 0;
}

static inline int check_for_channel_flood(int notice, aClient *sptr, aChannel *chptr)
{
	int delta = 0;

	if (!FloodConfig.msg_flood_count || sptr->user == NULL || chptr == NULL
	  || IsULine(sptr) || IsServer(sptr)) {
		return 0;
	}
	if (chptr->first_message_time + 1 < timeofday) {
		delta = timeofday - chptr->first_message_time;
		chptr->number_of_messages -= delta;
		chptr->first_message_time = timeofday;

		if (chptr->number_of_messages <= 0) {
			chptr->number_of_messages = 0;
			chptr->flood_notice = 0;
		}
	}
	if (chptr->number_of_messages >= FloodConfig.msg_flood_count || chptr->flood_notice) {
		if (!chptr->flood_notice) {
			sendto_realops_lev(FLOOD_LEV, "Possible flooder: %s (%s@%s) target: %s",
				sptr->name, sptr->username, MaskedHost(sptr), chptr->chname);
			chptr->flood_notice = 1;
			chptr->number_of_messages += 2;
		}
		if (MyClient(sptr) && !notice) {
			send_me_notice(sptr, ":Messages to %s throttled due to flooding",
				chptr->chname);
		}
		return 1;
	}
	chptr->number_of_messages++;
	return 0;
}

static inline int check_for_ctcp(char *str, char **dccptr)
{
	if (*str == 1) {
		if (!myncmp(&str[1], "DCC", 3)) {
			if (dccptr != NULL) {
				*dccptr = &str[1];
			}
			if (!myncmp(&str[4], " SEND", 5)) {
				return CTCP_DCCSEND;
			}
			return CTCP_DCC;
		}
		if (myncmp(&str[1], "ACTION", 6)) {
			return CTCP_YES;
		}
	}
	return CTCP_NONE;
}

static void send_error_msg(aClient *cptr, char *parv[], char *target, long send_errno)
{
	int err_numeric = ERR_CANNOTSENDTOCHAN;
	char *msg = "unknown cause";

	if (send_errno == CMODE_MODERATED) {
		msg = "channel is moderated";
	}
	else if (send_errno == CMODE_NOPRIVMSGS) {
		msg = "external channel messages not permitted";
	}
	else if (send_errno == CMODE_MODREGONLY) {
		err_numeric = ERR_NEEDREGGEDNICK;
		msg = "speak in";
	}
	else if (send_errno == CMODE_BAN) {
		msg = "you're banned";
	}
	else if (send_errno == CMODE_NOCOLOUR) {
		msg = parv[2];
		err_numeric = ERR_NOCOLOURSONCHAN;
	}
	else if (send_errno == CMODE_NOCTCP) {
		msg = "CTCP channel messages not permitted";
	}

	send_me_numeric(cptr, err_numeric, target, msg);
	return;
}

/*
 * parse_message
 *	parv[0] = sender prefix
 *	parv[1] = targets
 *	parv[2] = message
 */
static inline int parse_message(aClient *cptr, aClient *sptr, int parc, char *parv[], int notice, int directed)
{
	char *command = notice ? MSG_NOTICE : MSG_PRIVMSG;
	msg_ptr *msg = notice ? (directed ? &CMD_NOTICE_D : &CMD_NOTICE) : (directed ? &CMD_PRIVMSG_D : &CMD_PRIVMSG);
	char *nick, *p = NULL, *dccmsg, *s, *server;
	int i = 0, ret, ischan, ismine;
	aClient *acptr = NULL;
	aChannel *chptr;

	if (parc < 2 || *parv[1] == '\0') {
		send_me_numeric(sptr, ERR_NORECIPIENT, command);
		return 0;
	}
	if (parc < 3 || *parv[2] == '\0') {
		send_me_numeric(sptr, ERR_NOTEXTTOSEND, command);
		return 0;
	}

	ismine = MyClient(sptr);
	if (ismine) {
#if defined(ANTI_SPAMBOT) && !defined(ANTI_SPAMBOT_WARN_ONLY)
		if (sptr->localUser->join_part_count >= MAX_JOIN_LEAVE_COUNT) {
			if (mycmp(parv[0], parv[1])) {
				return 0;
			}
		}
#endif
		parv[1] = canonize(parv[1]);
	}

	for (nick = strtoken(&p, parv[1], ","); nick != NULL && i < 20;
	  nick = strtoken(&p, NULL, ",")) {
		if (ismine && (i++ > MAXTARGETS) && !IsULine(sptr)) {
			if ((FloodConfig.no_oper_flood && !HasMode(sptr, UMODE_OPER))
			  || !FloodConfig.no_oper_flood) {
				sptr->since += 4;
			}
		}

		ischan = IsChannelName(nick);
		if (ischan && ((chptr = find_channel(nick, NULL)) != NULL)) {
			if (!notice) {
				switch (check_for_ctcp(parv[2], NULL)) {
					case CTCP_NONE:
						break;
					case CTCP_DCCSEND:
					case CTCP_DCC:
						send_me_notice(sptr, ":You may not send DCC "
							"commands to a channel (%s)", nick);
						continue;
					default:
						if (chptr->mode.mode & CMODE_NOCTCP) {
							send_error_msg(sptr, parv, nick, CMODE_NOCTCP);
							continue;
						}
#ifdef FLUD
						if (check_for_flud(sptr, NULL, chptr)) {
							return 0;
						}
#endif
						break;
				}
			}

			ret = IsULine(sptr) ? 0 : can_send(sptr, chptr, parv[2]);
			if (ret) {
				if (!notice) {
					send_error_msg(sptr, parv, nick, ret);
				}
			}
			else {
				if (ismine) {
					if (check_for_channel_flood(notice, sptr, chptr)) {
						continue;
					}
					if (chptr->mode.mode & CMODE_STRIPCOLOUR) {
						parv[2] = strip_colour(parv[2]);
					}
				}
				sendto_channel_msg_butone_copy(cptr, sptr, chptr, nick, msg,
					PREFIX_COLON, parv[2]);
			}
			continue;
		}
		if (!ischan) {
			acptr = IsServer(cptr) ? find_person_target(nick) : find_person(nick, NULL);
			if (acptr != NULL) {
				HookData hdata;

				hdata.cptr = cptr;
				hdata.sptr = sptr;
				hdata.acptr = acptr;

				if (!notice && MyConnect(acptr)) {
					switch (check_for_ctcp(parv[2], &dccmsg)) {
						case CTCP_NONE:
							break;
						case CTCP_DCCSEND:
#ifdef FLUD
							if (check_for_flud(sptr, acptr, NULL)) {
								return 0;
							}
#endif

							hdata.c = dccmsg;
							if (hook_run_until(h_dccsend, &hdata, FLUSH_BUFFER) == FLUSH_BUFFER) {
								continue;
							}
							break;
						default:
#ifdef FLUD
							if (check_for_flud(sptr, acptr, NULL)) {
								return 0;
							}
#endif
							break;
					}
				}

				if (hook_run_until(h_privmsg, &hdata, FLUSH_BUFFER) == FLUSH_BUFFER) {
					continue;
				}
				if (ismine && check_for_client_flood(notice, sptr, acptr)) {
					continue;
				}
				if (!is_silenced(sptr, acptr)) {
					if (!notice && MyClient(acptr) && (acptr->user != NULL)
					  && (acptr->user->away != NULL)) {
						send_me_numeric(sptr, RPL_AWAY, acptr->name,
							acptr->user->away);
					}
					if (directed && !MyClient(acptr)) {
						sendto_one_client_onserver_copy(acptr, sptr, msg,
							PREFIX_COLON, parv[2]);
					}
					else {
						sendto_one_client_prefixed_copy(acptr, sptr, msg,
							PREFIX_COLON, parv[2]);
					}
				}
				continue;
			}
		}
		if ((s = strchr(nick, '#')) != NULL) {
			if ((chptr = find_channel(s, NULL)) == NULL) {
				target_left(sptr, s, command, parv[2]);
				continue;
			}
			if (ismine) {
				ret = IsULine(sptr) ? 0 : can_send(sptr, chptr, parv[2]);
				if (ret) {
					if (!notice) {
						send_error_msg(sptr, parv, nick, ret);
					}
					continue;
				}
				if (check_for_channel_flood(notice, sptr, chptr)) {
					continue;
				}
			}
			if (s == nick) {
				sendto_channel_msg_butone_copy(cptr, sptr, chptr, nick, msg,
					PREFIX_COLON, parv[2]);
			}
			else {
				char pfixed_chname[CHANNELLEN + 5];
				char *t, *x = pfixed_chname;
				int flags = 0;

				*x = '\0';
				for (t = nick; t != s; t++) {
					switch (*t) {
						case '+':
							if (!(flags & CMODE_VOICE)) {
								*x++ = '+';
							}
							flags |= CMODE_VOICE;
							break;
						case '%':
							if (!(flags & CMODE_HALFOP)) {
								*x++ = '%';
							}
							flags |= CMODE_HALFOP;
							break;
						case '@':
							if (!(flags & CMODE_CHANOP)) {
								*x++ = '@';
							}
							flags |= CMODE_CHANOP;
							break;
						case '*':
							if (!(flags & CMODE_CHANADMIN)) {
								*x++ = '*';
							}
							flags |= CMODE_CHANADMIN;
							break;
						default:
							break;
					}
				}
				for (t = s; *s != '\0'; s++) {
					*x++ = *s;
				}
				*x = '\0';
				sendto_chanflag_msg_butone_copy(cptr, sptr, chptr, pfixed_chname,
					flags, msg, PREFIX_COLON, parv[2]);
			}
			continue;
		}
		if (HasMode(sptr, UMODE_OPER|UMODE_SCLIENT)) {
			if (*nick == '$' || *nick == '#') {
				if ((s = strrchr(nick, '.')) == NULL) {
					send_me_numeric(sptr, ERR_NOTOPLEVEL, nick);
					continue;
				}
				while (*++s) {
					if (*s == '.' || *s == '*' || *s == '?') {
						break;
					}
				}
				if (*s == '*' || *s == '?') {
					send_me_numeric(sptr, ERR_WILDTOPLEVEL, nick);
					continue;
				}
				sendto_match_butone(IsServer(cptr) ? cptr : NULL, sptr, nick + 1,
					*nick == '#' ? MATCH_HOST : MATCH_SERVER, ":%s %s %s :%s", parv[0],
					command, nick, parv[2]);
				continue;
			}
		}
		if (!ischan && ((server = strchr(nick, '@')) != NULL) &&
		  ((acptr = find_server(server + 1, NULL)) != NULL)) {
			if (!IsMe(acptr)) {
				if (ismine) {
					aClient *ucptr;

					*server = '\0';
					ucptr = find_person(nick, NULL);
					*server = '@';

					if ((ucptr != NULL) && (ucptr->uplink == acptr)) {
						msg_ptr *msg_d = notice ? &CMD_NOTICE_D : &CMD_PRIVMSG_D;
						sendto_one_client_onserver_copy(ucptr, sptr, msg_d,
							PREFIX_COLON, parv[2]);
						continue;
					}
				}
				sendto_one_client_nopostfix(acptr, sptr, msg, "%s :%s", nick, parv[2]);
				continue;
			}

			*server = '\0';
			acptr = find_person(nick, NULL);
			*server = '@';

			if ((acptr != NULL) && !MyConnect(acptr)) {
				acptr = NULL;
			}
			if (acptr != NULL) {
				sendto_one_client_prefixed_copy(acptr, sptr, msg, PREFIX_COLON, parv[2]);
				continue;
			}
		}
		target_left(sptr, nick, command, parv[2]);
	}
	if (i > MAXTARGETS && sptr->user != NULL) {
		sendto_realops_lev(SPY_LEV, "%s (%s@%s) tried to message %d users", sptr->name,
			sptr->username, MaskedHost(sptr), i);
	}
	return 0;
}

/*
 * m_privmsg, m_notice
 *	parv[0] = sender prefix
 *	parv[1] = targets
 *	parv[2] = message
 */
int m_privmsg(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	return parse_message(cptr, sptr, parc, parv, 0, 0);
}
int m_privmsg_directed(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	return parse_message(cptr, sptr, parc, parv, 0, IsServer(cptr) ? 1 : 0);
}
int m_notice(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	return parse_message(cptr, sptr, parc, parv, 1, 0);
}
int m_notice_directed(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	return parse_message(cptr, sptr, parc, parv, 1, IsServer(cptr) ? 1 : 0);
}

