/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_mode.c: MODE command module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_mode.c,v 1.38 2004/05/13 17:43:51 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include "modules.h"
#include "xmode.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

Module MOD_HEADER(m_mode) = {
	"modules/commands/m_mode.so",
	"/MODE command",
	6, "$Revision: 1.38 $"
};

int MOD_LOAD(m_mode)()
{
	register_command(&MOD_HEADER(m_mode), &CMD_MODE, m_mode);
	MOD_SET_FLAG(&MOD_HEADER(m_mode), MOD_FLAG_PERM);
	return MOD_SUCCESS;
}

int MOD_UNLOAD(m_mode)()
{
	return MOD_SUCCESS;
}

static inline int get_privilege_level(aClient *sptr, aChannel *chptr)
{
	chanMember *cm = NULL;

	if (IsServer(sptr) && (chptr->channelts > 0)) {
		return CPRIV_SERVER;
	}
	if (IsULine(sptr) || (HasMode(sptr, UMODE_SADMIN) && !MyClient(sptr))) {
		return CPRIV_ULINE;
	}
	if ((cm = find_user_member(sptr->user->channel, chptr)) != NULL) {
		if (cm->flags & CMODE_CHANADMIN) {
			return CPRIV_CHANADMIN;
		}
		if (cm->flags & CMODE_CHANOP) {
			return CPRIV_CHANOP;
		}
		if (cm->flags & CMODE_HALFOP) {
			return CPRIV_HALFOP;
		}
		return CPRIV_NONE;
	}
	return -1;
}

int do_chan_mode(aClient *cptr, aClient *sptr, aChannel *chptr, int parc, char *parv[], int level)
{
	int count = parse_mode(cptr, sptr, chptr, level, parc, parv);

	Debug((DEBUG_DEBUG, "MODE(%s) modebuf [%s] parabuf [%s] idparabuf [%s]", chptr->chname,
		modebuf, parabuf, idparabuf));

	if (count < 1) {
		return 0;
	}
	if (*parabuf != '\0') {
		send_channel_msg_local(sptr, chptr, &CMD_MODE, "%s %s",
			modebuf, parabuf);
		sendto_serv_capab_msg_butone(cptr, sptr, NULL, 0, CAP_SUID|CAP_TOK1, &CMD_MODE,
			"%s %ld %s %s", chptr->chname, chptr->channelts, modebuf, parabuf);
		sendto_serv_capab_msg_butone(cptr, sptr, NULL, CAP_SUID|CAP_TOK1, 0, &CMD_MODE,
			"%s %ld %s %s", chptr->chname, chptr->channelts, modebuf, idparabuf);
	}
	else {
		send_channel_msg_local(sptr, chptr, &CMD_MODE, "%s", modebuf);
		sendto_serv_msg_butone(cptr, sptr, NULL, &CMD_MODE, "%s %ld %s", chptr->chname,
			chptr->channelts, modebuf);
	}
	return count;
}

/*
 * m_mode
 *	parv[0] = sender prefix
 *	parv[1] = channel
 */
int m_mode(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	int sub_parc = 2, level = CPRIV_NONE;
	aChannel *chptr;

	if (parc > 1) {
		if ((chptr = find_channel(parv[1], NULL)) == NULL) {
			return do_user_mode(cptr, sptr, parc, parv);
		}
	}
	else {
		send_me_numeric(sptr, ERR_NEEDMOREPARAMS, "MODE");
		return 0;
	}
	if (!check_channelname(sptr, parv[1])) {
		return 0;
	}

	if (parc < 3) {
		get_chan_modes(sptr, chptr, modebuf, parabuf);
		send_me_numeric(sptr, RPL_CHANNELMODEIS, chptr->chname, modebuf, parabuf);
		send_me_numeric(sptr, RPL_CREATIONTIME, chptr->chname, chptr->channelts);
		return 0;
	}

	if ((level = get_privilege_level(sptr, chptr)) == -1) {
		send_me_numeric(sptr, ERR_CHANOPRIVSNEEDED, chptr->chname);
		return 0;
	}

	if (IsServer(cptr) && isdigit(*parv[2])) {
		long mode_ts = 0;

		if ((mode_ts = atol(parv[2])) > 0 && (mode_ts > chptr->channelts)) {
			return 0;
		}
		sub_parc++;
	}

	do_chan_mode(cptr, sptr, chptr, parc - sub_parc, parv + sub_parc, level);
	return 0;
}
