/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_vhost.c: VHOST command/protocol module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_vhost.c,v 1.37 2004/04/03 19:55:00 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include "modules.h"
#include "xmode.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

Module MOD_HEADER(m_vhost) = {
	"modules/commands/m_vhost.so",
	"/VHOST command/protocol",
	5, "$Revision: 1.37 $"
};

int MOD_LOAD(m_vhost)()
{
	register_command(&MOD_HEADER(m_vhost), &CMD_VHOST, m_vhost);
	return MOD_SUCCESS;
}

int MOD_UNLOAD(m_vhost)()
{
	return MOD_SUCCESS;
}

static void handle_vhost_server(aClient *cptr, char *nick, char *mask)
{
	aClient *acptr;
	if (cptr == NULL || BadPtr(nick) || BadPtr(mask)) {
		ircdlog(LOG_ERROR, "handle_vhost_server() encountered a NULL pointer! cptr %lX "
			"nick %s mask %s", cptr, BadPtr(nick) ? "NULL!" : nick, BadPtr(mask) ? "NULL!" : mask);
		sendto_realops_lev(DEBUG_LEV, "Encountered NULL pointer in Server-VHOST function");
		return;
	}
	if ((acptr = find_target(nick)) == NULL) {
		ircdlog(LOG_ERROR, "Received Server-VHOST for unknown nick %s", nick);
		return;
	}

	AddMode(acptr, UMODE_MASKED);
	strncpyzt(acptr->user->maskedhost, mask, HOSTLEN + 1);
	sendto_serv_msg_butone(cptr, &me, NULL, &CMD_VHOST, "%s %s", acptr->name,
		acptr->user->maskedhost);
        return;
}

/*
 * m_vhost
 *	parv[0] = sender prefix
 *
 *	if (parc == 3) - server VHOST
 *		parv[1] = client
 *		parv[2] = vhost
 *
 *	if (parc == 2) - client VHOST OFF
 *		parv[1] = OFF
 *
 *	if (parc > 3) - client VHOST ON
 *		parv[1] = ON
 *		parv[2] = login
 *		parv[3] = password
 */
int m_vhost(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	if (IsServer(sptr)) {
		if (!CapVHOST(sptr)) {
			sendto_realops_lev(DEBUG_LEV, "Received VHOST command from non-VHOST "
				"server %s", get_client_name(sptr, HIDE_IP));
			return 0;
		}
		if (parc < 3 || BadPtr(parv[2])) {
			sendto_realops_lev(DEBUG_LEV, "Invalid parameter count in VHOST received "
				"from %s", get_client_name(sptr, HIDE_IP));
			return 0;
		}
		handle_vhost_server(sptr, parv[1], parv[2]);
		return 0;
	}

	if (!MyClient(sptr)) {
		return 0;
	}
	if (parc < 2 || (parc == 2 && BadPtr(parv[1])) || (parc > 3 && BadPtr(parv[3]))) {
		send_me_numeric(sptr, ERR_NEEDMOREPARAMS, "VHOST");
		return 0;
	}

	if (!mycmp(parv[1], "ON")) {
		dlink_node *node;
		ConfigItem_vhost *vhost = NULL;
		char *from;

		if (parc < 4) {
			send_me_numeric(cptr, ERR_NEEDMOREPARAMS, "VHOST");
			return 0;
		}

		from = make_user_host(cptr->username, cptr->host);
		DLINK_FOREACH_DATA(conf_vhost_list.head, node, vhost, ConfigItem_vhost) {
			if (!irccmp(vhost->login, parv[2]) && conf_check_from(&vhost->from, from)
			  && auth_check(vhost->auth, parv[3])) {
				break;
			}
		}
		if (vhost == NULL) {
			sendto_realops_lev(SPY_LEV, "%s (%s@%s) failed to log into a vhost",
				cptr->name, cptr->username, MaskedHost(cptr));
			send_me_noticeNA(cptr, ":Authentication failure.");
			return 0;
		}

		sendto_realops_lev(SPY_LEV, "%s (%s@%s) has successfully logged into vhost %s",
			cptr->name, cptr->username, MaskedHost(cptr), vhost->mask);
		if (!HasMode(cptr, UMODE_MASKED)) {
			AddMode(cptr, UMODE_MASKED);
			sendto_one_client(cptr, cptr, &CMD_MODE, ":+x");
		}

		strncpyzt(cptr->user->maskedhost, vhost->mask, HOSTLEN + 1);
		SetVhost(cptr);

		send_me_notice(cptr, ":*** Your hostname is now: %s", cptr->user->maskedhost);
		sendto_serv_msg_butone(cptr, &me, NULL, &CMD_VHOST, "%s %s", cptr->name,
			cptr->user->maskedhost);
	}
	else if (!mycmp(parv[1], "OFF")) {
		if (!HasVhost(cptr)) {
			return 0;
		}
		if (!HasMode(cptr, UMODE_MASKED)) {
			AddMode(cptr, UMODE_MASKED);
			sendto_one_client(cptr, cptr, &CMD_MODE, ":+x");
		}

		ClearVhost(cptr);
		strncpyzt(cptr->user->maskedhost, maskme(cptr), HOSTLEN + 1);

		send_me_notice(cptr, ":*** Your hostname is now: %s", cptr->user->maskedhost);
		sendto_serv_msg_butone(NULL, &me, NULL, &CMD_VHOST, "%s %s", cptr->name,
			cptr->user->maskedhost);
	}
	else {
		send_me_notice(cptr, ":*** VHOST %s: Unknown command", parv[1]);
	}

	return 0;
}
