/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_whois.c: WHOIS command module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_whois.c,v 1.46 2004/05/13 17:43:51 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include "hook.h"
#include "modules.h"
#include "xmode.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

static char buf[BUFSIZE];
static int bufpos = 0;

Module MOD_HEADER(m_whois) = {
	"modules/commands/m_whois.so",
	"/WHOIS command",
	6, "$Revision: 1.46 $"
};

int MOD_LOAD(m_whois)()
{
	register_command(&MOD_HEADER(m_whois), &CMD_WHOIS, m_whois);
	return MOD_SUCCESS;
}

int MOD_UNLOAD(m_whois)()
{
	return MOD_SUCCESS;
}

static inline void w_addtobuf(char *string)
{
	char *ptr;
	for (ptr = string; *ptr; ptr++) {
		buf[bufpos++] = *ptr;
	}
	buf[bufpos] = '\0';
	return;
}

/*
 * m_whois
 *	parv[0] = sender prefix
 *	parv[1] = nickname list
 */
int m_whois(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	aClient *acptr;
	aChannel *chptr;
	chanMember *cm;
	char *nick, *tmp, *name = NULL, *p = NULL;
	int found, len, showchan = 0;
	static time_t last_used = 0L;

	if (!HasMode(sptr, UMODE_OPER)) {
		if (FloodConfig.pace_wait_simple && (last_used + FloodConfig.pace_wait_simple > timeofday)) {
			send_me_numericNA(sptr, RPL_LOAD2HI);
			return 0;
		}
		else {
			last_used = timeofday;
		}
	}
	if (parc < 2 || *parv[1] == '\0') {
		send_me_numeric(sptr, ERR_NEEDMOREPARAMS, "WHOIS");
		return 0;
	}
	if (parc > 2) {
		if (use_or_deliver(cptr, sptr, &CMD_WHOIS, "%s :%s", 1, parc, parv) != HUNTED_ISME) {
			return 0;
		}
		parv[1] = parv[2];
	}
	for (tmp = parv[1]; (nick = strtoken(&p, tmp, ",")); tmp = NULL) {
		found = 0;
		buf[(bufpos = 0)] = '\0';

		if ((acptr = find_client(nick, NULL)) == NULL || !IsPerson(acptr)) {
			target_left(sptr, nick, MSG_WHOIS, NULL);
			return 0;
		}
		name = *acptr->name != '\0' ? acptr->name : "?";

		send_me_numeric(sptr, RPL_WHOISUSER, name, acptr->username, MaskedHost(acptr),
			acptr->info);
		if (HasMode(sptr, UMODE_OPER) || sptr == acptr) {
			send_me_numeric(sptr, RPL_WHOISHOSTANDMODES, name, acptr->host,
				get_user_modes(acptr));
		}
		if (!HasMode(acptr, UMODE_SCLIENT)) {
			len = strlen(me.name) + strlen(parv[0]) + 6 + strlen(name);
			for (cm = acptr->user->channel; cm != NULL; cm = cm->nextchan) {
				chptr = cm->chptr;
				showchan = ShowChannel(sptr, chptr) ? 1 : 0;

				if (!showchan && !HasMode(sptr, UMODE_SADMIN)) {
					continue;
				}
				if (bufpos + strlen(chptr->chname) > BUFSIZE - 4 - len) {
					send_me_numeric(sptr, RPL_WHOISCHANNELS, name, buf);
					buf[(bufpos = 0)] = '\0';
				}
				if (bufpos > 0) {
					buf[bufpos++] = ' ';
				}
				if (showchan == 0) {
					buf[bufpos++] = '!';
				}
				if (cm->flags & CMODE_CHANADMIN) {
					buf[bufpos++] = '*';
				}
				else if (cm->flags & CMODE_CHANOP) {
					buf[bufpos++] = '@';
				}
				else if (cm->flags & CMODE_HALFOP) {
					buf[bufpos++] = '%';
				}
				else if (cm->flags & CMODE_VOICE) {
					buf[bufpos++] = '+';
				}
				w_addtobuf(chptr->chname);
			}
			if (bufpos > 0) {
				send_me_numeric(sptr, RPL_WHOISCHANNELS, name, buf);
				buf[(bufpos = 0)] = '\0';
			}
		}

		send_me_numeric(sptr, RPL_WHOISSERVER, name, acptr->user->server,
			acptr->uplink != NULL ? acptr->uplink->info : "*NOT ON THIS NET*");
		if (HasMode(acptr, UMODE_REGNICK)) {
			send_me_numeric(sptr, RPL_WHOISREGNICK, name);
		}
		if (acptr->user->away != NULL) {
			send_me_numeric(sptr, RPL_AWAY, name, acptr->user->away);
		}
		if (HasMode(acptr, UMODE_OPER)) {
			char *extra = NULL;
			if (HasMode(acptr, UMODE_NETADMIN)) {
				extra = " - Network Administrator";
			}
			else if (HasMode(acptr, UMODE_ADMIN)) {
				extra = " - Server Administrator";
			}
			send_me_numeric(sptr, RPL_WHOISOPERATOR, name,
				(extra != NULL) ? extra : "");
		}
		if (HasMode(acptr, UMODE_SADMIN)) {
			send_me_numeric(sptr, RPL_WHOISSADMIN, name);
		}
		if (HasMode(acptr, UMODE_SCLIENT)) {
			send_me_numeric(sptr, RPL_WHOISSERVICECLI, name);
		}
		if (IsSecure(acptr)) {
			send_me_numeric(sptr, RPL_WHOISSECURE, name);
		}
		if (MyConnect(acptr)) {
			ASSERT(acptr->localUser != NULL);
			send_me_numeric(sptr, RPL_WHOISIDLE, name,
				(timeofday - acptr->localUser->last), acptr->firsttime);
		}
		continue;
	}
	send_me_numeric(sptr, RPL_ENDOFWHOIS, name);
	return 0;
}
