/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/adns.c: libadns interface functions
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: adns.c,v 1.23 2004/05/13 20:17:53 rageircd Exp $
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "res.h"
#include "numeric.h"
#include "h.h"
#include "memory.h"
#include "fd.h"
#include "internal.h"
#include "adns.h"
#include "internal.h"
#include <stdlib.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/socket.h>

adns_state ircd_dns;

void dns_delete_queries(DNSQuery *q)
{
	if (q != NULL && (q->query != NULL)) {
		adns_cancel(q->query);
	}
}

static void dns_io()
{
	struct adns_pollfd pollfds[MAXFD_POLL], *pfd;
	int pfd_cnt, i;

	adns__consistency(ircd_dns, 0, cc_entex);
	pfd_cnt = adns__pollfds(ircd_dns, pollfds);

	for (i = 0; i < pfd_cnt; i++) {
		pfd = &pollfds[i];
		if (pfd->events & ADNS_POLLIN) {
			engine_set_call(pfd->fd, FDL_SERVER, FD_WANT_READ, dns_read,
				NULL, 0);
		}
		if (pfd->events & ADNS_POLLOUT) {
			engine_set_call(pfd->fd, FDL_SERVICE, FD_WANT_WRITE, dns_write,
				NULL, 0);
		}
	}
}

static void adns_timeout()
{
	adns_processtimeouts(ircd_dns, &Internal.ircd_time);
}

int init_resolver()
{
	int i;

	i = adns_init(&ircd_dns, adns_if_noautosys, 0);

	if (ircd_dns == NULL) {
		return 0;
	}

	add_event_approx("adns_timeout", adns_timeout, NULL, 5, 1);
	dns_io();

	return 1;
}

void restart_resolver()
{
	adns__rereadconfig(ircd_dns);
}

static void dns_process_callbacks()
{
	adns_query query, r;
	adns_answer *answer;
	DNSQuery *q;
	int i;

	adns_forallqueries_begin(ircd_dns);

	while ((query = adns_forallqueries_next(ircd_dns, (void **)&r)) != NULL) {
		i = adns_check(ircd_dns, &query, &answer, (void **)&q);

		if (i != EAGAIN) {
			ASSERT(q->callback != NULL);
			q->query = NULL;
			q->callback(q->ptr, (!i) ? answer : NULL);
		}
	}
}

void dns_read(int fd, void *unused)
{
	adns_processreadable(ircd_dns, fd, &Internal.ircd_time);
	dns_io();
	dns_process_callbacks();
}

void dns_write(int fd, void *unused)
{
	adns_processwriteable(ircd_dns, fd, &Internal.ircd_time);
	dns_io();
	dns_process_callbacks();
}

int dns_gethost(const char *name, DNSQuery *q)
{
	int i;

	ASSERT(ircd_dns->nservers > 0);

	i = adns_submit(ircd_dns, name, adns_r_addr, adns_qf_owner, q, &q->query);
	dns_io();

	return i;
}

int dns_getaddr(struct in_addr *ip, DNSQuery *q)
{
	struct sockaddr_in addr;
	int i;

	ASSERT(ircd_dns->nservers > 0);

	memset(&addr, '\0', sizeof(struct sockaddr_in));
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = ip->s_addr;
	addr.sin_port = 0;

	i = adns_submit_reverse(ircd_dns, (struct sockaddr *)&addr, adns_r_ptr,
		adns_qf_owner|adns_qf_cname_loose|adns_qf_quoteok_anshost, q, &q->query);
	dns_io();

	return i;
}

void report_dns_nameservers(aClient *sptr)
{
	int i;

	send_me_noticeNA(sptr, ":*** Current DNS name servers:");
	for (i = 0; i < ircd_dns->nservers; i++) {
		send_me_notice(sptr, ":*** %s", inetntoa((const char *)&ircd_dns->servers[i].addr));
	}
}
