/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/channel_ban.c: channel ban list management
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: channel_ban.c,v 1.8 2004/02/18 03:33:48 rageircd Exp $
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include "hook.h"
#include "xmode.h"

BlockHeap *channelban_heap;

void init_channelbans()
{
	channelban_heap = BlockHeapCreate((size_t)sizeof(channelBan), CHANNELBANS_PREALLOCATE);
	return;
}

static dlink_list *get_ban_list(aChannel *chptr, long type)
{
	if (type == CMODE_BAN) {
		return &chptr->banlist;
	}
	if (type == CMODE_EXCEPT) {
		return &chptr->exceptlist;
	}
	if (type == CMODE_INVEX) {
		return &chptr->invexlist;
	}

	ASSERT(1 == 0);
	return NULL;
}

static char get_ban_flag(long type)
{
	if (type == CMODE_BAN) {
		return 'b';
	}
	if (type == CMODE_EXCEPT) {
		return 'e';
	}
	if (type == CMODE_INVEX) {
		return 'I';
	}

	ASSERT(1 == 0);
	return '?';
}

static int match_id(char *id, aClient *acptr)
{
	ASSERT(acptr != NULL);

	if (HasMode(acptr, UMODE_MASKED) && irccmp(acptr->user->maskedhost, acptr->host)) {
		if (!match(id, make_nick_user_host(acptr->name, acptr->username, acptr->user->maskedhost))) {
			return 1;
		}
	}
	if (!match(id, make_nick_user_host(acptr->name, acptr->username, acptr->host))) {
		return 1;
	}
	if (!match(id, make_nick_user_host(acptr->name, acptr->username, acptr->hostip))) {
		return 1;
	}
	return 0;
}

void destroy_channelban(channelBan *ban)
{
	ASSERT(ban != NULL);
	MyFree(ban->id);
	MyFree(ban->set_by);
	free_channelban(ban);
	return;
}

void send_ban_list(aClient *sptr, aChannel *chptr, dlink_list *list, int listn, int endn)
{
	dlink_node *node;
	channelBan *ban;

	ASSERT(sptr != NULL);
	ASSERT(chptr != NULL);
	ASSERT(list != NULL);

	DLINK_FOREACH_DATA(list->head, node, ban, channelBan) {
		send_me_numeric(sptr, listn, chptr->chname, ban->id, ban->set_by, ban->when);
	}
	send_me_numeric(sptr, endn, chptr->chname);
	return;
}

void synch_ban_list(aClient *sptr, aChannel *chptr, long mode)
{
	dlink_list *list;
	dlink_node *node;
	char flag;
	channelBan *ban;
	int n = 0, send_it = 0;

	ASSERT(sptr != NULL);
	ASSERT(chptr != NULL);

	flag = get_ban_flag(mode);
	list = get_ban_list(chptr, mode);

	midx = pidx = 0;
	modebuf[0] = parabuf[0] = '\0';
	ADD_MODE('+');

	DLINK_FOREACH_DATA(list->head, node, ban, channelBan) {
		if ((pidx + strlen(ban->id) + 10) < MODEBUFLEN) {
			ADD_MODE(flag);
			ADD_NPARA(ban->id);
			n++;
		}
		else if (pidx) {
			send_it = 1;
		}
		if (n == MAXTSMODEPARAMS) {
			send_it = 1;
		}
		if (send_it) {
			sendto_one_client_nopostfix(sptr, &me, &CMD_MODE, "%s %ld %s %s",
				chptr->chname, chptr->channelts, modebuf, parabuf);

			midx = pidx = send_it = 0;
			modebuf[0] = parabuf[0] = '\0';
			ADD_MODE('+');

			if (n != MAXTSMODEPARAMS) {
				ADD_MODE(flag);
				ADD_NPARA(ban->id);
				n = 1;
				continue;
			}
			n = 0;
		}
	}
	if (midx > 1) {
		sendto_one_client_nopostfix(sptr, &me, &CMD_MODE, "%s %ld %s %s", chptr->chname,
			chptr->channelts, modebuf, parabuf);
	}
	return;
}

void kill_ban_list(aClient *sptr, aChannel *chptr, long mode)
{
	dlink_list *list;
	dlink_node *node, *next = NULL;
	char flag;
	channelBan *ban;
	chanMember *cm;
	int n = 0, send_it = 0;

	ASSERT(sptr != NULL);
	ASSERT(chptr != NULL);

	flag = get_ban_flag(mode);
	list = get_ban_list(chptr, mode);

	midx = pidx = 0;
	modebuf[0] = parabuf[0] = '\0';
	ADD_MODE('-');

	DLINK_FOREACH_SAFE_DATA(list->head, node, next, ban, channelBan) {
		if ((pidx + strlen(ban->id) + 10) < MODEBUFLEN) {
			ADD_MODE(flag);
			ADD_NPARA(ban->id);
			n++;
		}
		else if (pidx) {
			send_it = 1;
		}
		if (n >= MAXMODEPARAMS) {
			send_it = 1;
		}

		if (send_it) {
			send_channel_msg_local(&me, chptr, &CMD_MODE, "%s %s", modebuf, parabuf);
			send_it = 0;

			midx = pidx = 0;
			modebuf[0] = parabuf[0] = '\0';
			ADD_MODE('-');

			if (n != MAXMODEPARAMS) {
				ADD_MODE(flag);
				ADD_NPARA(ban->id);
				n = 1;
				continue;
			}

			n = 0;
		}

		dlink_del(list, NULL, node);
		destroy_channelban(ban);
	}
	if (midx > 1) {
		send_channel_msg_local(&me, chptr, &CMD_MODE, "%s %s", modebuf, parabuf);
	}

	for (cm = chptr->members; cm != NULL; cm = cm->nextuser) {
		if (MyClient(cm->cptr)) {
			cm->bans = 0;
		}
	}

	return;
}

int add_id(char *id, aChannel *chptr, long type, aClient *cptr)
{
	dlink_list *list;
	dlink_node *node;
	channelBan *ban;

	ASSERT(!BadPtr(id));
	ASSERT(chptr != NULL);
	ASSERT(cptr != NULL);

	list = get_ban_list(chptr, type);

	if (MyClient(cptr) && (dlink_length(list) >= MAXBANS)) {
		send_me_numeric(cptr, ERR_BANLISTFULL, chptr->chname, id);
		return 0;
	}
	DLINK_FOREACH_DATA(list->head, node, ban, channelBan) {
		if (!match(ban->id, id)) {
			return 0;
		}
	}

	ban = make_channelban();
	ban->when = timeofday;
	DupString(ban->id, id);

	if (IsPerson(cptr)) {
		ban->set_by = (char *)MyMallocX(strlen(cptr->name) + strlen(cptr->username) +
			strlen(MaskedHost(cptr)) + 3);
		ircsprintf(ban->set_by, "%s!%s@%s", cptr->name, cptr->username, MaskedHost(cptr));
	}
	else {
		DupString(ban->set_by, cptr->name);
	}
	dlink_add(list, ban);

	if (type == CMODE_BAN) {
		chanMember *cm;
		for (cm = chptr->members; cm != NULL; cm = cm->nextuser) {
			if (MyConnect(cm->cptr) && match_id(id, cm->cptr)) {
				cm->bans++;
			}
		}
	}
	return 1;
}

int del_id(char *id, aChannel *chptr, long type)
{
	dlink_list *list;
	dlink_node *node = NULL, *next = NULL;
	channelBan *ban;
	chanMember *cm = NULL;

	ASSERT(!BadPtr(id));
	ASSERT(chptr != NULL);

	list = get_ban_list(chptr, type);
	DLINK_FOREACH_SAFE_DATA(list->head, node, next, ban, channelBan) {
		if (mycmp(id, ban->id)) {
			continue;
		}

		dlink_del(list, NULL, node);
		destroy_channelban(ban);

		if (type != CMODE_BAN) {
			return 1;
		}

		for (cm = chptr->members; cm != NULL; cm = cm->nextuser) {
			if (cm->bans && MyConnect(cm->cptr) && match_id(id, cm->cptr)) {
				cm->bans--;
			}
		}
		return 1;
	}
	return 0;
}

int del_id_bynode(dlink_node *node, aChannel *chptr, long type)
{
	dlink_list *list;
	channelBan *ban;

	ASSERT(node != NULL);
	ASSERT(chptr != NULL);

	list = get_ban_list(chptr, type);
	ban = (channelBan *)node->data;

	dlink_del(list, NULL, node);

	if (type == CMODE_BAN) {
		chanMember *cm;
		for (cm = chptr->members; cm != NULL; cm = cm->nextuser) {
			if (cm->bans && MyConnect(cm->cptr) && match_id(ban->id, cm->cptr)) {
				cm->bans--;
			}
		}
	}

	destroy_channelban(ban);
	return 1;
}

int has_invex(aClient *cptr, aChannel *chptr)
{
	dlink_node *node;
	channelBan *invex;

	if (!IsPerson(cptr)) {
		return 0;
	}

	DLINK_FOREACH_DATA(chptr->invexlist.head, node, invex, channelBan) {
		if (match_id(invex->id, cptr)) {
			return 1;
		}
	}
	return 0;
}

int has_except(aClient *cptr, aChannel *chptr)
{
	dlink_node *node;
	channelBan *except;

	if (!IsPerson(cptr)) {
		return 0;
	}

	DLINK_FOREACH_DATA(chptr->exceptlist.head, node, except, channelBan) {
		if (match_id(except->id, cptr)) {
			return 1;
		}
	}
	return 0;
}

int has_ban(aClient *cptr, aChannel *chptr)
{
	dlink_node *node;
	channelBan *ban;

	if (!IsPerson(cptr)) {
		return 0;
	}

	DLINK_FOREACH_DATA(chptr->banlist.head, node, ban, channelBan) {
		if (match_id(ban->id, cptr) && !has_except(cptr, chptr)) {
			return 1;
		}
	}
	return 0;
}
