/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/engine.c: socket engine functions
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: engine.c,v 1.40 2004/05/13 22:11:46 rageircd Exp $
 */

#include "setup.h"
#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "sock.h"
#include "h.h"
#include "fd.h"
#include "zlink.h"
#include "memory.h"
#include "res.h"
#include "ssl.h"
#include "hook.h"
#include "xmode.h"
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>

static int flush_recvqs(HookData *);
static HookEvent *he_flush_recvqs;

#ifdef INCREASE_SOCK_BUFS
extern char *readbuf;
extern int rcvbufmax, sndbufmax;
#else
extern char readbuf[MAX_CLIENT_RECVQ];
#endif

static const char *engine_errtab[] = {
	"Engine OK",
	"Error during bind()",
	"Error during DNS lookup",
	"Connect timeout",
	"Error during connect()",
	"Engine error"
};

static char *engine_get_sockerr(aClient *cptr)
{
	ASSERT(cptr != NULL);
	ASSERT(cptr->localClient != NULL);

	if (!cptr->localClient->sockerr) {
		return "No error";
	}
	if (cptr->localClient->sockerr > 0) {
		return strerror(cptr->localClient->sockerr);
	}

	switch (cptr->localClient->sockerr) {
		case IRCERR_BUFALLOC:
			return "sbuf allocation error";
		case IRCERR_ZIP:
			return "compression failure";
		case IRCERR_SSL:
			return "SSL error";
		default:
			break;
	}

	return "unknown error";
}

static int engine_parse_client_queue(aClient *cptr)
{
	int dolen = 0, done;

#ifdef INCREASE_SOCK_BUFS
	size_t readbufsize = rcvbufmax * sizeof(char);
#else
	size_t readbufsize = sizeof(readbuf);
#endif

	ASSERT(cptr != NULL);
	ASSERT(cptr->localClient != NULL);

	while (SBufLength(&cptr->localClient->recvQ) && !NoNewLine(cptr) && ((cptr->status < STAT_UNKNOWN) ||
	  (cptr->since - timeofday < 10) || DoingDKEY(cptr))) {
		if (IsServer(cptr) || DoingDKEY(cptr)) {
			if ((dolen = sbuf_get(&cptr->localClient->recvQ, readbuf, readbufsize)) <= 0) {
				break;
			}
			if ((done = serv_dopacket(cptr, readbuf, dolen))) {
				return done;
			}
			break;
		}
		if ((dolen = sbuf_getmsg(&cptr->localClient->recvQ, readbuf, readbufsize)) <= 0) {
			if (dolen < 0) {
				return exit_client(cptr, cptr, cptr, "sbuf_getmsg() fail");
			}
			if (SBufLength(&cptr->localClient->recvQ) < 510) {
				SetNoNewLine(cptr);
			}
			else {
				SBufClear(&cptr->localClient->recvQ);
			}
			break;
		}
		else if (user_dopacket(cptr, readbuf, dolen) == FLUSH_BUFFER) {
			return FLUSH_BUFFER;
		}
		else if (DeadSocket(cptr)) {
			break;
		}
	}
	if (SBufLength(&cptr->localClient->recvQ) && !NoNewLine(cptr) && !ParsingRecvQ(cptr)) {
		SetParsingRecvQ(cptr);
		dlink_add(&recvqcli_list, cptr);
	}
	return 0;
}

static void engine_read_error(aClient *cptr, int recv_len)
{
	char error_buf[512];
	int err_no;
	char *error_str;

	ASSERT(cptr != NULL);
	ASSERT(cptr->localClient != NULL);

	err_no = get_sockerr(cptr->localClient->fd);
	error_str = strerror(err_no);

	Debug((DEBUG_ERROR, "READ ERROR: fd = %d recv %d err_no %d (%s)", cptr->localClient->fd,
		recv_len, err_no, error_str));

	if (IsServer(cptr) || IsHandshake(cptr) || IsConnecting(cptr)) {
		char *path = get_client_name(cptr, HIDE_IP);
		if (!recv_len) {
			sendto_gnotice("Server %s closed the connection.", path);
			sendto_serv_msg_butone(cptr, &me, NULL, &CMD_GNOTICE, ":Server %s "
				"closed the connection.", path);
		}
		else {
			if (IsConnecting(cptr) || IsHandshake(cptr)) {
				sendto_gnotice("Connect error to %s (%s)", path, error_str);
				sendto_serv_msg_butone(cptr, &me, NULL, &CMD_GNOTICE,
					"Connect error to %s (%s)", path, error_str);
			}
			else {
				sendto_gnotice("Read error from %s (%s), closing link", path, error_str);
				sendto_serv_msg_butone(cptr, &me, NULL, &CMD_GNOTICE,
					":Read error from %s (%s), closing link", path, error_str);
			}
		}
	}
	if (!recv_len) {
		ircsprintf(error_buf, "Client closed connection");
	}
	else {
		ircsprintf(error_buf, "Read error: %s", error_str);
	}

	exit_client(cptr, cptr, &me, error_buf);
}

void engine_write_error(aClient *cptr, char *error, int sockerr)
{
	char *errmsg, errbuf[512];

	ASSERT(cptr != NULL);
	ASSERT(cptr->localClient != NULL);

	cptr->localClient->sockerr = sockerr;
	SetDeadSocket(cptr);

	SBufClear(&cptr->localClient->recvQ);
	SBufClear(&cptr->localClient->sendQ);

	errmsg = SendQEx(cptr) ? "SendQ exceeded" : engine_get_sockerr(cptr);
	if (!IsPerson(cptr) && !IsUnknown(cptr) && !IsClosing(cptr)) {
		char *path = get_client_name(cptr, HIDE_IP);

		ircsprintf(errbuf, "%s", error);
		sendto_gnotice(errbuf, path, errmsg);
		sendto_serv_msg_butone(cptr, &me, NULL, &CMD_GNOTICE, errbuf, path, errmsg);
	}

	ircsprintf(errbuf, "Write error: %s", errmsg);
	exit_client(cptr, cptr, &me, errbuf);
}

void engine_connect_error(aClient *cptr, char *error, int err_no)
{
	char *errmsg, errbuf[512];

	ASSERT(cptr != NULL);
	ASSERT(cptr->localClient != NULL);

	errmsg = strerror(err_no);

	if (IsConnecting(cptr) && !BadPtr(error)) {
		sendto_gnotice(error, get_client_name(cptr, SHOW_IP), errmsg);
	}

	ircsprintf(errbuf, "Connect error: %s", errmsg);
	exit_client(cptr, cptr, &me, errbuf);
}

int ircsend(aClient *cptr, const void *ptr, size_t size)
{
	ASSERT(cptr != NULL);
	ASSERT(cptr->localClient != NULL);
#ifdef USE_OPENSSL
	if (IsSecure(cptr) && (cptr->localClient->ssl != NULL)) {
		return safe_SSL_write(cptr, ptr, size);
	}
#endif
	return send(cptr->localClient->fd, ptr, size, 0);
}

int ircrecv(aClient *cptr, void *ptr, size_t size)
{
	ASSERT(cptr != NULL);
	ASSERT(cptr->localClient != NULL);
#ifdef USE_OPENSSL
	if (IsSecure(cptr) && (cptr->localClient->ssl != NULL)) {
		return safe_SSL_read(cptr, ptr, size);
	}
#endif
	return recv(cptr->localClient->fd, ptr, size, 0);
}

static int engine_write(aClient *cptr, char *buf, int len)
{
	int send_len;

	ASSERT(cptr != NULL);
	ASSERT(cptr->localClient != NULL);

#ifdef DEBUGMODE
	Internal.debug.write_calls++;
#endif

	if ((send_len = ircsend(cptr, buf, len)) < 0 && engine_ignore_errno(errno)) {
#ifdef DEBUGMODE
		Internal.debug.write_errors_ignored++;
#endif
		SetBlocked(cptr);
		return 0;
	}
#ifdef DEBUGMODE
	else if (send_len <= 0) {
		Internal.debug.write_errors++;
	}
#endif
	if (send_len > 0) {
		if (IsBlocked(cptr)) {
			ClearBlocked(cptr);
		}

		cptr->localClient->sendB += send_len;
		if (cptr->localClient->sendB & 0x0400) {
			cptr->localClient->sendK += (cptr->localClient->sendB >> 10);
			cptr->localClient->sendB &= 0x03ff;
		}

		me.localClient->sendB += send_len;
		if (me.localClient->sendB & 0x0400) {
			me.localClient->sendK += (me.localClient->sendB >> 10);
			me.localClient->sendB &= 0x03ff;
		}

		if (cptr->localClient->listener != NULL) {
			cptr->localClient->listener->sendB += send_len;
			if (cptr->localClient->listener->sendB & 0x0400) {
				cptr->localClient->listener->sendK += (cptr->localClient->listener->sendB >> 10);
				cptr->localClient->listener->sendB &= 0x03ff;
			}
		}
	}
	return send_len;
}

void engine_send_queued(int fd, void *data)
{
	aClient *cptr = (aClient *)data;
	int send_len, write_len, more_data = 0;
	char *send_buf;

	ASSERT(cptr != NULL);
	ASSERT(cptr->localClient != NULL);

#ifdef USE_OPENSSL
	if (IsSecure(cptr) && (cptr->localClient->ssl != NULL) && !SSL_is_init_finished(cptr->localClient->ssl)) {
		if (safe_SSL_accept(cptr->localClient->ssl, cptr->localClient->fd, cptr->localClient->sockhost) == -1
		  || DeadSocket(cptr)) {
			engine_write_error(cptr, "SSL initialisation failure", IRCERR_SSL);
		}
		else {
			engine_set_call(cptr->localClient->fd, FDL_CLIENT, FD_WANT_WRITE,
				engine_send_queued, cptr, 0);
		}
		return;
	}
#endif
	if (DeadSocket(cptr)) {
		ircdlog(LOG_ERROR, "engine_send_queued() called for dead client %s[%d][%s]",
			cptr->name, cptr->localClient->fd, cptr->from->name);
		return;
	}

	if (IsServer(cptr) && OutputZIP(cptr) && zip_is_data_out(cptr->serv->zip_out)) {
		if (SBufLength(&cptr->localClient->sendQ)) {
			more_data = 1;
		}
		else {
			int ldata = SendingBurst(cptr);

			send_buf = zip_output(cptr->serv->zip_out, NULL, &send_len, 1, &ldata);
			if (send_len == -1) {
				sendto_realops("Zip output error for %s: (%d) %s", cptr->name,
					ldata, send_buf);
				engine_write_error(cptr, "Zip output error for %s", IRCERR_ZIP);
				return;
			}

#ifdef USE_OPENSSL
			if (OutputRC4(cptr)) {
				rc4_process_stream_to_buf(cptr->serv->rc4_out, send_buf, rc4buf, send_len);
				send_buf = rc4buf;
			}
#endif

			if (!sbuf_put(&cptr->localClient->sendQ, send_buf, send_len)) {
				engine_write_error(cptr, "Buffer allocation error for %s", IRCERR_BUFALLOC);
				return;
			}
		}
	}

	while (SBufLength(&cptr->localClient->sendQ)) {
		send_buf = sbuf_map(&cptr->localClient->sendQ, &send_len);

		if ((write_len = engine_write(cptr, send_buf, send_len)) < 0) {
			engine_write_error(cptr, "Write error to %s, closing link (%s)", errno);
			return;
		}

		sbuf_delete(&cptr->localClient->sendQ, write_len);
		cptr->localClient->last_sendq = (SBufLength(&cptr->localClient->sendQ) >> 10);

		if (send_len > write_len) {
			break;
		}
		if (more_data && !SBufLength(&cptr->localClient->sendQ)) {
			int ldata = SendingBurst(cptr);

			more_data = 0;
			send_buf = zip_output(cptr->serv->zip_out, NULL, &send_len, 1, &ldata);
			if (send_len == -1) {
				sendto_realops("Zip output error (more) for %s: (%d) %s",
					cptr->name, ldata, send_buf);
				engine_write_error(cptr, "Zip output error (more) for %s",
					IRCERR_ZIP);
				return;
			}

#ifdef USE_OPENSSL
			if (OutputRC4(cptr)) {
				rc4_process_stream_to_buf(cptr->serv->rc4_out, send_buf, rc4buf, send_len);
				send_buf = rc4buf;
			}
#endif

			if (!sbuf_put(&cptr->localClient->sendQ, send_buf, send_len)) {
				engine_write_error(cptr, "Buffer allocation error (more) for %s",
					IRCERR_BUFALLOC);
				return;
			}
		}
	}

	if (IsServer(cptr) && CapBURST(cptr) && SentSOB(cptr) && (SBufLength(&cptr->localClient->sendQ) < 20480)) {
		if (!SendingBurst(cptr)) {
			ClearSentSOB(cptr);
			sendto_one_client_nopostfix(cptr, NULL, &CMD_BURST, "%d",
				SBufLength(&cptr->localClient->sendQ));
		}
	}

	if (SBufLength(&cptr->localClient->sendQ)) {
		engine_set_call(fd, FDL_CLIENT, FD_WANT_WRITE, engine_send_queued, cptr, 0);
	}
}

void engine_read_packet(int fd, void *data)
{
	aClient *cptr = (aClient *)data;
	int recv_len = 0;
#ifdef INCREASE_SOCK_BUFS
	size_t readbufsize = IsPerson(cptr) ? (MAX_CLIENT_RECVQ * sizeof(char)) : (rcvbufmax * sizeof(char));
#else
	size_t readbufsize = sizeof(readbuf);
#endif

	ASSERT(cptr != NULL);
	ASSERT(cptr->localClient != NULL);

#ifdef USE_OPENSSL
	if (IsSecure(cptr) && (cptr->localClient->ssl != NULL) && !SSL_is_init_finished(cptr->localClient->ssl)) {
		if (safe_SSL_accept(cptr->localClient->ssl, cptr->localClient->fd, cptr->localClient->sockhost) == -1
		  || DeadSocket(cptr)) {
			engine_read_error(cptr, 0);
		}
		else {
			engine_set_call(cptr->localClient->fd, FDL_CLIENT, FD_WANT_READ,
				engine_read_packet, cptr, 0);
		}
		return;
	}
#endif
	if (DeadSocket(cptr)) {
		return;
	}

	if (!(IsPerson(cptr) && SBufLength(&cptr->localClient->recvQ) > MAX_CLIENT_RECVQ)) {
		errno = 0;

#ifdef DEBUGMODE
		Internal.debug.read_calls++;
#endif

		if ((recv_len = ircrecv(cptr, readbuf, readbufsize)) <= 0) {
			if (recv_len == -1 && engine_ignore_errno(errno)) {
#ifdef DEBUGMODE
				Internal.debug.read_errors_ignored++;
#endif

				engine_set_call(cptr->localClient->fd, FDL_CLIENT, FD_WANT_READ,
					engine_read_packet, cptr, 0);
				return;
			}

#ifdef DEBUGMODE
			Internal.debug.read_errors++;
#endif
			engine_read_error(cptr, recv_len);
			return;
		}

		ClearPingSent(cptr);
		ClearNoNewLine(cptr);

		if ((cptr->lasttime = timeofday) > cptr->since) {
			cptr->since = cptr->lasttime;
		}
	}
	if (IsServer(cptr) || IsConnecting(cptr) || IsHandshake(cptr) || DoingDKEY(cptr)) {
		if (recv_len > 0) {
			if (!serv_dopacket(cptr, readbuf, recv_len)) {
				engine_set_call(cptr->localClient->fd, FDL_SERVER, FD_WANT_READ,
					engine_read_packet, cptr, 0);
			}
			return;
		}
	}
	else {
		if (sbuf_put(&cptr->localClient->recvQ, readbuf, recv_len) < 0) {
			exit_client(cptr, cptr, cptr, "sbuf_put() fail");
			return;
		}
		if (IsPerson(cptr) && ((FloodConfig.no_oper_flood && !HasMode(cptr, UMODE_OPER)) ||
		  !FloodConfig.no_oper_flood) && (SBufLength(&cptr->localClient->recvQ) > CLIENT_FLOOD)) {
			sendto_realops_lev(FLOOD_LEV, "%s!%s@%s (%d) Exceeded %d RecvQ",
				cptr->name, cptr->username, MaskedHost(cptr),
				SBufLength(&cptr->localClient->recvQ), CLIENT_FLOOD);
			exit_client(cptr, cptr, cptr, "Excess Flood");
			return;
		}
		engine_parse_client_queue(cptr);
	}
	if (!DeadSocket(cptr)) {
		engine_set_call(cptr->localClient->fd, FDL_CLIENT, FD_WANT_READ, engine_read_packet, cptr, 0);
	}
}

int engine_ignore_errno(int errtmp)
{
	switch (errtmp) {
		case EALREADY:
		case EINTR:
		case EINPROGRESS:
		case EWOULDBLOCK:
#if EAGAIN != EWOULDBLOCK
		case EAGAIN:
#endif
#ifdef ERESTART
		case ERESTART:
#endif
			return 1;
		default:
			break;
	}
	return 0;
}

const char *engine_strerror(int error)
{
	if (error < 0 || error >= ENGINE_ERR_MAX) {
		return "Invalid error number!";
	}
	return engine_errtab[error];
}

int engine_open(int sock_type, int proto, char *desc)
{
	int fd;

	if (fds.count >= MASTER_MAX) {
		errno = ENFILE;
		return -1;
	}
	if ((fd = socket(AF_INET, sock_type, proto)) < 0) {
		return -1;
	}
	if (!set_non_blocking(fd)) {
		ircdlog(LOG_ERROR, "engine_open() fd %d set_non_blocking() failure", fd);
		close(fd);
		return -1;
	}

	fd_open(fd, desc);
	return fd;
}

int engine_accept(int fd, struct sockaddr_in *addr)
{
	int newfd, addrlen = sizeof(struct sockaddr_in);

	if (fds.count >= MASTER_MAX) {
		errno = ENFILE;
		return -1;
	}
	if ((newfd = accept(fd, (struct sockaddr *)addr, &addrlen)) == -1) {
		return -1;
	}
	if (!set_non_blocking(newfd)) {
		ircdlog(LOG_ERROR, "engine_accept() fd %d set_non_blocking() failure", newfd);
		close(newfd);
		return -1;
	}

	fd_open(newfd, "client connection");
	return newfd;
}

void engine_set_timeout(int fd, time_t timeout, void (*h)(), void *d)
{
	fd_entry *fde = fd_get_entry(fd);

	FD_USED(fd)

	fde->timeout = timeofday + (timeout / 1000);
	fde->h_timeout = h;
	fde->d_timeout = d;
}

void engine_set_flush(int fd, time_t timeout, void (*h)(), void *d)
{
	fd_entry *fde = fd_get_entry(fd);

	FD_USED(fd)

	fde->flush_timeout = timeofday + (timeout / 1000);
	fde->h_flush = h;
	fde->d_flush = d;
}

void engine_check_timeouts()
{
	int fd;
	fd_entry *fde;
	void (*h)();
	void *d;

	for (fd = 0; fd <= fds.highest; fd++) {
		fde = fd_get_entry(fd);
		if (!fde->open) {
			continue;
		}

		if (fde->h_flush != NULL && (fde->flush_timeout > 0) && (timeofday > fde->flush_timeout)) {
			h = fde->h_flush;
			d = fde->d_flush;

			engine_set_flush(fd, 0, NULL, NULL);
			h(fd, d);
		}
		if (fde->h_timeout != NULL && (fde->timeout > 0) && (timeofday > fde->timeout)) {
			h = fde->h_timeout;
			d = fde->d_timeout;

			engine_set_timeout(fd, 0, NULL, NULL);
			h(fd, d);
		}
	}
}

static void engine_connect_callback(int fd, int status)
{
	fd_entry *fde = fd_get_entry(fd);
	void (*h)();

	FD_USED(fd)

	h = fde->h_connect;
	fde->h_connect = NULL;

	if (h != NULL) {
		engine_set_timeout(fd, 0, NULL, NULL);
		h(fd, status, fde->d_connect);
	}
}

static void engine_connect_try(int fd, void *data)
{
	fd_entry *fde = fd_get_entry(fd);
	int retval;

	FD_USED(fd)

	ASSERT(fde->h_connect != NULL);

	if ((retval = connect(fd, (struct sockaddr *)&fde->ip, sizeof(struct sockaddr))) < 0) {
		if (errno == EISCONN) {
			engine_connect_callback(fd, ENGINE_OK);
		}
		else if (engine_ignore_errno(errno)) {
			engine_set_call(fd, FDL_SERVER, FD_WANT_WRITE, engine_connect_try, NULL, 0);
		}
		else {
			engine_connect_callback(fd, ENGINE_ERR_CONNECT);
		}
		return;
	}

	engine_connect_callback(fd, ENGINE_OK);
}

static void engine_connect_timeout(int fd, void *data)
{
	engine_connect_callback(fd, ENGINE_ERR_TIMEOUT);
}

static void engine_connect_dns_callback(void *vfd, adns_answer *answer)
{
	fd_entry *fde = (fd_entry *)vfd;

	if (answer == NULL) {
		engine_connect_callback(fde->fd, ENGINE_ERR_DNS);
		return;
	}
	if (answer->status != adns_s_ok) {
		engine_connect_callback(fde->fd, ENGINE_ERR_DNS);
		MyFree(answer);
		free_dnsquery(fde->dns_query);
		return;
	}

	fde->ip.sin_addr.s_addr = answer->rrs.addr->addr.inet.sin_addr.s_addr;
	engine_set_timeout(fde->fd, 30 * 1000, engine_connect_timeout, NULL);

	MyFree(answer);
	free_dnsquery(fde->dns_query);

	engine_connect_try(fde->fd, NULL);
}

void engine_connect_tcp(int fd, const char *host, int port, struct sockaddr *sa, int salen,
  void (*h)(), void *d, int timeout)
{
	fd_entry *fde = fd_get_entry(fd);

	FD_USED(fd)

	fde->h_connect = h;
	fde->d_connect = d;
	fde->ip.sin_family = AF_INET;
	fde->ip.sin_port = htons(port);

	if (sa != NULL && (bind(fd, sa, salen) < 0)) {
		engine_connect_callback(fd, ENGINE_ERR_BIND);
		return;
	}

	if (!inetpton(host, &fde->ip.sin_addr.s_addr)) {
		fde->dns_query = make_dnsquery();
		fde->dns_query->ptr = fde;
		fde->dns_query->callback = engine_connect_dns_callback;
		dns_gethost(host, fde->dns_query);
	}
	else {
		engine_set_timeout(fd, timeout * 1000, engine_connect_timeout, NULL);
		engine_connect_try(fd, NULL);
	}
}

static int flush_recvqs(HookData *hdata)
{
	dlink_node *node, *next = NULL;
	aClient *cptr;

	DLINK_FOREACH_SAFE_DATA(recvqcli_list.head, node, next, cptr, aClient) {
		ASSERT(cptr->localClient != NULL);

		if (SBufLength(&cptr->localClient->recvQ) && !NoNewLine(cptr)) {
			if (engine_parse_client_queue(cptr) == FLUSH_BUFFER) {
				continue;
			}
		}
		if (!(SBufLength(&cptr->localClient->recvQ) && !NoNewLine(cptr))) {
			dlink_del(&recvqcli_list, NULL, node);
			ClearParsingRecvQ(cptr);
		}
	}
	return 0;
}

void engine_preinit()
{
	he_flush_recvqs = hook_add_event(h_pre_netio, flush_recvqs);
}
