/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/engine_kqueue.c: kqueue socket engine
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: engine_kqueue.c,v 1.18 2004/04/10 15:22:48 rageircd Exp $
 */

#include "setup.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "h.h"
#include "fd.h"
#include "memory.h"
#include <sys/types.h>
#include <sys/time.h>

#ifdef USE_KQUEUE

#include <sys/event.h>

#define KEVENT_LENGTH 128

#ifndef EV_SET
#define EV_SET(kep, a, b, c, d, e, f)	do { \
						(kep)->ident = (a); \
						(kep)->filter = (b); \
						(kep)->flags = (c); \
						(kep)->fflags = (d); \
						(kep)->data = (e); \
						(kep)->udata = (f); \
					} while (0)
#endif

typedef struct _ircd_kqueue ircd_kqueue;
struct _ircd_kqueue {
	int id;
	struct kevent *list;
	int max;
	int offset;
	struct timespec timespec;
};

static ircd_kqueue kq;

static void kq_update_events(int fd, short filter, void (*h)())
{
	fd_entry *fde = fd_get_entry(fd);
	void (*curr_h)();
	int kep_flags = EV_DELETE;

	switch (filter) {
		case EVFILT_READ:
			curr_h = fde->h_read;
			break;
		case EVFILT_WRITE:
			curr_h = fde->h_write;
			break;
		default:
			return;
			break;
	}

	if ((curr_h == NULL && (h != NULL)) || (curr_h != NULL && (h == NULL))) {
		struct kevent *kep = kq.list + kq.offset;

		if (h != NULL) {
			kep_flags = EV_ADD;
			if (filter == EVFILT_WRITE) {
				kep_flags |= EV_ONESHOT;
			}
		}

		EV_SET(kep, fd, filter, kep_flags, 0, 0, 0);

		if (kq.offset == kq.max) {
			int ret = -1;
			if ((ret = kevent(kq.id, kq.list, kq.offset, NULL, 0, &kq.timespec)) < 0) {
				ircdlog(LOG_ERROR, "kevent() returned error %d: %s", errno,
					strerror(errno));
				return;
			}
			kq.offset = 0;
		}
		else {
			kq.offset++;
		}
	}
}
				
void engine_init()
{
	memset(&kq, '\0', sizeof(ircd_kqueue));
	if ((kq.id = kqueue()) < 0) {
		ircdlog(LOG_ERROR, "FATAL: engine_init() with kqueue failure %d: %s",
			errno, strerror(errno));
		abort();
	}

	kq.max = getdtablesize();
	kq.list = (struct kevent *)MyMalloc(sizeof(struct kevent) * kq.max);
	kq.timespec.tv_sec = 0;
	kq.timespec.tv_nsec = 0;
}

void engine_set_call(int fd, fd_list list, unsigned int flags, void (*h)(), void *d, time_t timeout)
{
	fd_entry *fde = fd_get_entry(fd);
	FD_USED(fd)

	fde->list = list;
	fde->flags |= flags;

	if (flags & FD_WANT_READ) {
		kq_update_events(fd, EVFILT_READ, h);
		fde->h_read = h;
		fde->d_read = d;
	}
	if (flags & FD_WANT_WRITE) {
		kq_update_events(fd, EVFILT_WRITE, h);
		fde->h_write = h;
		fde->d_write = d;
	}
	if (timeout) {
		engine_set_timeout(fd, timeout, NULL, NULL);
	}
}

void engine_do_netio(time_t delay)
{
	int cnt, i, fd;
	static struct kevent ke[KEVENT_LENGTH];
	struct timespec ts;
	fd_entry *fde;
	void (*h)();

	do {
		ts.tv_sec = 0;
		ts.tv_nsec = (delay * 1000);

		for (;;) {
			cnt = kevent(kq.id, kq.list, kq.offset, ke, KEVENT_LENGTH, &ts);
			kq.offset = 0;

			if (cnt >= 0) {
				break;
			}
			if (!engine_ignore_errno(errno)) {
				cnt = 0;
				break;
			}
		}

		set_time();

		if (!cnt) {
			return;
		}

		for (i = 0; i < cnt; i++) {
			fd = (int)ke[i].ident;
			fde = fd_get_entry(fd);

			if (ke[i].flags & EV_ERROR) {
				errno = ke[i].data;
				continue;
			}

			switch (ke[i].filter) {
				case EVFILT_READ:
					if ((h = fde->h_read) != NULL) {
						fde->flags &= ~FD_WANT_READ;
						fde->h_read = NULL;
						h(fd, fde->d_read);
					}
				case EVFILT_WRITE:
					if ((h = fde->h_write) != NULL) {
						fde->flags &= ~FD_WANT_WRITE;
						fde->h_write = NULL;
						h(fd, fde->d_write);
					}
				default:
					break;
			}
		}
		return;
	} while (0);
	return;
}

#endif
