/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/engine_poll.c: poll socket engine
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: engine_poll.c,v 1.20 2004/04/10 15:22:48 rageircd Exp $
 */

#include "setup.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "h.h"
#include "fd.h"

#ifdef USE_POLL

#include POLLH

#ifndef POLLRDNORM
#define POLLRDNORM POLLIN
#endif
#ifndef POLLWRNORM
#define POLLWRNORM POLLOUT
#endif

typedef struct _pollfd_list pollfd_list;
struct _pollfd_list {
	struct pollfd list[MAXCONNECTIONS];
	int max_index;
};

static pollfd_list poll_fds;

static int poll_find_slot()
{
	int i;

	for (i = 0; i < MAXCONNECTIONS; i++) {
		if (poll_fds.list[i].fd == -1) {
			return i;
		}
	}

	ASSERT("poll_find_slot() with no slots" == NULL);

	return -1;
}

static void poll_update_fds(int fd, short event, void (*h)())
{
	fd_entry *fde = fd_get_entry(fd);
	struct pollfd *pfd;
	int i;

	if (fde->engine_index < 0) {
		fde->engine_index = poll_find_slot();
	}

	i = fde->engine_index;
	pfd = &poll_fds.list[i];

	if (h != NULL) {
		fde->list = FDL_ACTIVE;
		pfd->events |= event;
		pfd->fd = fd;

		if (i > poll_fds.max_index) {
			poll_fds.max_index = i;
		}
	}
	else if (i >= 0) {
		pfd->events &= ~event;
		if (pfd->events) {
			return;
		}

		pfd->fd = -1;
		pfd->revents = 0;

		fde->list = FDL_CLIENT;
		fde->flags = 0;
		fde->engine_index = -1;

		if (i == poll_fds.max_index) {
			while (poll_fds.max_index >= 0 && (poll_fds.list[poll_fds.max_index].fd == -1)) {
				poll_fds.max_index--;
			}
		}
	}
}

void engine_init()
{
	int i;

	for (i = 0; i < MAXCONNECTIONS; i++) {
		poll_fds.list[i].fd = -1;
		poll_fds.list[i].events = 0;
		poll_fds.list[i].revents = 0;
	}
	poll_fds.max_index = 0;
}

void engine_set_call(int fd, fd_list list, unsigned int flags, void (*h)(), void *d, time_t timeout)
{
	fd_entry *fde = fd_get_entry(fd);
	FD_USED(fd)

	fde->list = list;
	fde->flags |= flags;

	if (flags & FD_WANT_READ) {
		fde->h_read = h;
		fde->d_read = d;
		poll_update_fds(fd, POLLRDNORM, h);
	}
	if (flags & FD_WANT_WRITE) {
		fde->h_write = h;
		fde->d_write = d;
		poll_update_fds(fd, POLLWRNORM, h);
	}
	if (timeout) {
		engine_set_timeout(fd, timeout, NULL, NULL);
	}
}

void engine_do_netio(time_t delay)
{
	int cnt, i, fd;
	struct pollfd *pfd;
	fd_entry *fde;
	void (*h)();

	for (;;) {
		if ((cnt = poll(poll_fds.list, poll_fds.max_index + 1, delay)) >= 0) {
			break;
		}
		if (!engine_ignore_errno(errno)) {
			cnt = 0;
			break;
		}
	}

	set_time();

	if (!cnt) {
		return;
	}

	for (i = 0; i < poll_fds.max_index + 1; i++) {
		pfd = &poll_fds.list[i];

		if (!pfd->revents || pfd->fd == -1) {
			continue;
		}

		fd = pfd->fd;
		fde = fd_get_entry(fd);

		if (pfd->revents & (POLLRDNORM|POLLIN|POLLHUP|POLLERR)) {
			poll_update_fds(fd, POLLRDNORM, NULL);
			if ((h = fde->h_read) != NULL) {
				fde->h_read = NULL;
				h(fd, fde->d_read);
			}
		}
		if (pfd->revents & (POLLWRNORM|POLLOUT|POLLHUP|POLLERR)) {
			poll_update_fds(fd, POLLWRNORM, NULL);
			if ((h = fde->h_write) != NULL) {
				fde->h_write = NULL;
				h(fd, fde->d_write);
			}
		}
	}
}

#endif
