/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/engine_rtsigio.c: Real-Time SIGIO socket engine
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: engine_rtsigio.c,v 1.10 2004/04/10 23:40:01 rageircd Exp $
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE 1
#endif

#include "setup.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "h.h"
#include "fd.h"

#ifdef USE_RTSIGIO

#include <signal.h>
#include <unistd.h>
#include <fcntl.h>
#include POLLH

#define SIGIO_SIGNAL SIGRTMIN

#ifndef POLLRDNORM
#define POLLRDNORM POLLIN
#endif
#ifndef POLLWRNORM
#define POLLWRNORM POLLOUT
#endif

typedef struct _pollfd_list pollfd_list;
struct _pollfd_list {
	struct pollfd list[MAXCONNECTIONS];
	int max_index;
};

static pollfd_list poll_fds;

static int my_pid;
static int queue_overflow;
static sigset_t sigio_set;

static void mask_our_signal()
{
	sigemptyset(&sigio_set);
	sigaddset(&sigio_set, SIGIO_SIGNAL);
	sigaddset(&sigio_set, SIGIO);
	sigprocmask(SIG_BLOCK, &sigio_set, NULL);
}

static int poll_find_slot()
{
	int i;

	for (i = 0; i < MAXCONNECTIONS; i++) {
		if (poll_fds.list[i].fd == -1) {
			return i;
		}
	}

	ASSERT("poll_find_slot() with no slots" == NULL);
	return -1;
}

static void poll_update_fds(int fd, short event, void (*h)())
{
	fd_entry *fde = fd_get_entry(fd);
	struct pollfd *pfd;
	int i;

	if (fde->engine_index < 0) {
		fde->engine_index = poll_find_slot();
	}

	i = fde->engine_index;
	pfd = &poll_fds.list[i];

	if (h != NULL) {
		fde->list = FDL_ACTIVE;
		pfd->events |= event;
		pfd->fd = fd;

		if (i > poll_fds.max_index) {
			poll_fds.max_index = i;
		}
	}
	else if (i >= 0) {
		pfd->events &= ~event;
		if (pfd->events) {
			return;
		}

		pfd->fd = -1;
		pfd->revents = 0;

		fde->flags = 0;
		fde->list = FDL_CLIENT;
		fde->engine_index = -1;

		if (i == poll_fds.max_index) {
			while (poll_fds.max_index >= 0 && (poll_fds.list[poll_fds.max_index].fd == -1)) {
				poll_fds.max_index--;
			}
		}
	}
}

void setup_sigio_fd(int fd)
{
	fcntl(fd, F_SETFL, fcntl(fd, F_GETFL)|O_ASYNC|O_NONBLOCK);
	fcntl(fd, F_SETSIG, SIGIO_SIGNAL);
	fcntl(fd, F_SETOWN, my_pid);
}

void engine_init()
{
	int i;

	for (i = 0; i < MAXCONNECTIONS; i++) {
		poll_fds.list[i].fd = -1;
		poll_fds.list[i].events = 0;
		poll_fds.list[i].revents = 0;
	}
	poll_fds.max_index = 0;

	my_pid = getpid();
	mask_our_signal();
}

void engine_set_call(int fd, fd_list list, unsigned int flags, void (*h)(), void *d, time_t timeout)
{
	fd_entry *fde = fd_get_entry(fd);
	FD_USED(fd)

	fde->list = list;
	fde->flags |= flags;

	if (flags & FD_WANT_READ) {
		fde->h_read = h;
		fde->d_read = d;
		poll_update_fds(fd, POLLRDNORM, h);
	}
	if (flags & FD_WANT_WRITE) {
		fde->h_write = h;
		fde->d_write = d;
		poll_update_fds(fd, POLLWRNORM, h);
	}
	if (timeout) {
		engine_set_timeout(fd, timeout, NULL, NULL);
	}
}

void engine_do_netio(time_t delay)
{
	int cnt = 0, i, fd, sig;
	struct pollfd *pfd;
	struct siginfo si;
	struct timespec ts;
	fd_entry *fde;
	void (*h)();

	ts.tv_sec = 0;
	ts.tv_nsec = (delay * 1000);

	for (;;) {
		if (queue_overflow) {
			break;
		}
		if ((sig = sigtimedwait(&sigio_set, &si, &ts)) <= 0) {
			break;
		}

		if (sig == SIGIO) {
			ircdlog(LOG_ERROR, "Kernel RT signal queue overflow. Please check "
				"/proc/sys/kernel/rtsig-max is big enough!");
			queue_overflow = 1;
			break;
		}

		set_time();

		fd = si.si_fd;
		fde = fd_get_entry(fd);

		pfd = &poll_fds.list[fde->engine_index];
		pfd->revents |= si.si_band;

		if (pfd->revents & (POLLRDNORM|POLLIN|POLLHUP|POLLERR)) {
			poll_update_fds(fd, POLLRDNORM, NULL);
			if ((h = fde->h_read) != NULL) {
				fde->h_read = NULL;
				h(fd, fde->d_read);
			}
		}
		if (pfd->revents & (POLLWRNORM|POLLOUT|POLLHUP|POLLERR)) {
			poll_update_fds(fd, POLLWRNORM, NULL);
			if ((h = fde->h_write) != NULL) {
				fde->h_write = NULL;
				h(fd, fde->d_write);
			}
		}
	}

	if (!queue_overflow) {
		set_time();
		return;
	}

	signal(SIGIO_SIGNAL, SIG_IGN);
	signal(SIGIO_SIGNAL, SIG_DFL);
	queue_overflow = 0;

	for (;;) {
		if ((cnt = poll(poll_fds.list, poll_fds.max_index + 1, 0)) >= 0) {
			break;
		}
		if (!engine_ignore_errno(errno)) {
			cnt = 0;
			break;
		}
	}

	set_time();

	if (!cnt) {
		return;
	}

	for (i = 0; i < poll_fds.max_index + 1; i++) {
		pfd = &poll_fds.list[i];
		if (!pfd->revents || pfd->fd == -1) {
			continue;
		}

		fd = pfd->fd;
		fde = fd_get_entry(fd);

		if (pfd->revents & (POLLRDNORM|POLLIN|POLLHUP|POLLERR)) {
			poll_update_fds(fd, POLLRDNORM, NULL);
			if ((h = fde->h_read) != NULL) {
				fde->h_read = NULL;
				h(fd, fde->d_read);
			}
		}
		if (pfd->revents & (POLLWRNORM|POLLOUT|POLLHUP|POLLERR)) {
			poll_update_fds(fd, POLLWRNORM, NULL);
			if ((h = fde->h_write) != NULL) {
				fde->h_write = NULL;
				h(fd, fde->d_write);
			}
		}
	}

	mask_our_signal();
}

#endif
