/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/event.c: internal events management
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: event.c,v 1.20 2004/04/09 13:12:28 rageircd Exp $
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "patchlevel.h"
#include <time.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include "h.h"
#include "memory.h"

dlink_list event_list;

#ifdef STATIC_MODULES
extern void expire_klines();
#endif

static dlink_node *find_event_node(void (*func)(), void *arg)
{
	dlink_node *node;
	Event *event;

	DLINK_FOREACH_DATA(event_list.head, node, event, Event) {
		if (event->func == func && (event->arg == arg)) {
			return node;
		}
	}
	return NULL;
}

Event *add_event(char *name, void (*func)(), void *arg, time_t when, int loop)
{
	Event *event = (Event *)MyMallocX(sizeof(Event));
	DupString(event->name, name);
	event->func = func;
	event->arg = arg;
	event->when = (timeofday + when);
	event->freq = when;
	event->loop = (loop) ? 1 : 0;
	dlink_add(&event_list, event);

	if (Internal.next_event > when || (Internal.next_event == -1)) {
		Internal.next_event = when;
	}

	return event;
}

Event *add_event_approx(char *name, void (*func)(), void *arg, time_t when, int loop)
{
	if (when > 3) {
		const time_t two_thirds = (2 * when) / 3;
		when = two_thirds + ((rand() % 1000) * two_thirds) / 1000;
	}

	return add_event(name, func, arg, when, loop);
}

void del_event(Event *event, dlink_node *node)
{
	ASSERT(event != NULL);

	dlink_del(&event_list, event, node);

	MyFree(event->name);
	MyFree(event);
	return;
}

void del_event_byfunc(void (*func)(), void *arg)
{
	dlink_node *node;
	if ((node = find_event_node(func, arg)) == NULL) {
		return;
	}
	del_event((Event *)node->data, node);
	return;
}

void run_events()
{
	dlink_node *node, *next = NULL;
	Event *event;

	DLINK_FOREACH_SAFE_DATA(event_list.head, node, next, event, Event) {
		if (event->when > timeofday) {
			continue;
		}

		event->func(event->arg);

		if (event->loop) {
			event->when = (timeofday + event->freq);
		}
		else {
			del_event(event, node);
		}
		Internal.next_event = -1;
	}
	return;
}

void get_next_event()
{
	dlink_node *node;
	Event *event;

	if (Internal.next_event != -1) {
		return;
	}
	DLINK_FOREACH_DATA(event_list.head, node, event, Event) {
		if (Internal.next_event > event->when || (Internal.next_event == -1)) {
			Internal.next_event = event->when;
		}
	}
	return;
}

void show_events(aClient *sptr)
{
	dlink_node *node;
	Event *event;

	if (!dlink_length(&event_list)) {
		send_me_noticeNA(sptr, ":*** There are no pending events");
		return;
	}

	DLINK_FOREACH_DATA(event_list.head, node, event, Event) {
		send_me_notice(sptr, ":*** %s (type: %s, next execution: %s)", event->name,
			event->loop ? "continuous" : "once", time_to_str(event->when - timeofday));
	}
	send_me_notice(sptr, ":*** There are %d pending event(s)", dlink_length(&event_list));

	return;
}
