/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/fd.c: fd management
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: fd.c,v 1.15 2004/05/10 16:08:06 rageircd Exp $
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "h.h"
#include "fd.h"
#include "memory.h"
#include "numeric.h"

fd_entry fd_table[MAXCONNECTIONS];
fd_stat fds;

static void fd_init_slot(fd_entry *fde)
{
	fde->fd = -2;
	fde->engine_index = -1;
	fde->list = FDL_NONE;
	fde->open = 0;
	fde->flags = 0;

	fde->h_read = NULL;
	fde->d_read = NULL;

	fde->h_write = NULL;
	fde->d_write = NULL;

	fde->timeout = 0;
	fde->h_timeout = NULL;
	fde->d_timeout = NULL;

	fde->flush_timeout = 0;
	fde->h_flush = NULL;
	fde->d_flush = NULL;
}

static void fd_check_highest(int fd, int opening)
{
	if (fds.highest > fd) {
		return;
	}

	if (fd > fds.highest) {
		ASSERT(opening);
		fds.highest = fd;
	}
	else {
		ASSERT(!opening);
		while (fds.highest >= 0 && !fd_table[fds.highest].open) {
			fds.highest--;
		}
	}
}

void fd_init()
{
	int i;

	memset(fd_table, '\0', sizeof(fd_entry) * MAXCONNECTIONS);
	for (i = 0; i < MAXCONNECTIONS; i++) {
		fd_init_slot(&fd_table[i]);
	}

	memset(&fds, '\0', sizeof(fd_stat));
	fds.highest = -1;
	fds.count = 0;
}

void fd_open(int fd, char *desc)
{
	fd_entry *fde = fd_get_entry(fd);
	FD_UNUSED(fd)

	fde->fd = fd;
	fde->open = 1;
	strncpyzt(fde->desc, desc, 64);

	fd_check_highest(fd, 1);
	fds.count++;
}

void fd_close(int fd)
{
	fd_entry *fde = fd_get_entry(fd);
	FD_USED(fd)

	engine_set_call(fd, FDL_NONE, FD_WANT_READ|FD_WANT_WRITE, NULL, NULL, 0);

	fde->open = 0;
	*fde->desc = '\0';

	fd_check_highest(fd, 0);
	fds.count--;

	fd_init_slot(fde);
	close(fd);
}

#ifdef DEBUGMODE

static char *fd_get_list(fd_list list)
{
	switch (list) {
		case FDL_NONE:
			return "none";
		case FDL_ACTIVE:
			return "active";
		case FDL_SERVER:
			return "server";
		case FDL_SERVICE:
			return "service";
		case FDL_CLIENT:
			return "client";
		default:
			break;
	}
	return "unknown";
}

static char *fd_get_state(unsigned short flags)
{
	switch (flags) {
		case 3:
			return "read/write";
		case 2:
			return "write";
		case 1:
			return "read";
		case 0:
			return "idle";
		default:
			break;
	}
	return "unknown(?)";
}

void fd_report(aClient *cptr)
{
	int i;
	fd_entry *fde;

	send_me_noticeNA(cptr, ":*** File Descriptor Statistics");
	for (i = 0; i < MAXCONNECTIONS; i++) {
		fde = fd_get_entry(i);
		if (!fde->open) {
			continue;
		}
		send_me_notice(cptr, ":Open socket %d (type: %s, state: %s) [%s]", fde->fd,
			fd_get_list(fde->list), fd_get_state(fde->flags), fde->desc);
	}

	send_me_notice(cptr, ":*** Total: %d Highest: %d", fds.count, fds.highest);
}

#endif
