/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/hook.c: internal function hooks
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: hook.c,v 1.35 2004/05/13 16:12:54 rageircd Exp $
 */

#include "struct.h"
#include "setup.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "h.h"
#include "msg.h"
#include "channel.h"
#include "send.h"
#include "patchlevel.h"
#include "memory.h"
#include "hook.h"
#include "dlink.h"
#include "modules.h"
#include <sys/types.h>
#include <sys/stat.h>   
#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

dlink_list hook_list;

Hook *h_pre_access = NULL;
Hook *h_post_access = NULL;
Hook *h_post_motd = NULL;
Hook *h_conf_parse = NULL;
Hook *h_conf_test = NULL;
Hook *h_conf_verify = NULL;
Hook *h_conf_complete = NULL;
Hook *h_conf_rehash = NULL;
Hook *h_pre_netio = NULL;
Hook *h_post_netio = NULL;
Hook *h_exit_user_local = NULL;
Hook *h_exit_user_remote = NULL;
Hook *h_exit_server = NULL;
Hook *h_exit_unknown = NULL;
Hook *h_pre_connauth = NULL;
Hook *h_post_register_user = NULL;
Hook *h_post_oper = NULL;

void hook_init()
{
	h_pre_access = hook_add("h_pre_access");
	h_post_access = hook_add("h_post_access");
	h_post_motd = hook_add("h_post_motd");
	h_conf_parse = hook_add("h_conf_parse");
	h_conf_test = hook_add("h_conf_test");
	h_conf_verify = hook_add("h_conf_verify");
	h_conf_complete = hook_add("h_conf_complete");
	h_conf_rehash = hook_add("h_conf_rehash");
	h_pre_netio = hook_add("h_pre_netio");
	h_post_netio = hook_add("h_post_netio");
	h_exit_user_local = hook_add("h_exit_user_local");
	h_exit_user_remote = hook_add("h_exit_user_remote");
	h_exit_server = hook_add("h_exit_server");
	h_exit_unknown = hook_add("h_exit_unknown");
	h_pre_connauth = hook_add("h_pre_connauth");
	h_post_register_user = hook_add("h_post_register_user");
	h_post_oper = hook_add("h_post_oper");
}

static void *hook_find_event(Hook *hook, int (*event)(), int want_node)
{
	dlink_node *node;
	HookEvent *hookevent;

	ASSERT(hook != NULL);

	DLINK_FOREACH_DATA(hook->events.head, node, hookevent, HookEvent) {
		if (hookevent->event == event) {
			return (want_node) ? (void *)node : (void *)hookevent;
		}
	}

	return NULL;
}

Hook *hook_find_from_module(char *name, char *owner)
{
	Hook *hook = NULL;

	ASSERT(!BadPtr(name));
	ASSERT(!BadPtr(owner));

	if ((hook = hook_find(name)) == NULL) {
		if (do_module("load", owner, NULL) == MOD_SUCCESS) {
			hook = hook_find(name);
		}
	}

	return hook;
}

Hook *hook_find(char *name)
{
	dlink_node *node;
	Hook *hook;

	DLINK_FOREACH_DATA(hook_list.head, node, hook, Hook) {
		if (!irccmp(hook->name, name)) {
			return hook;
		}
	}

	return NULL;
}

HookEvent *hook_add_event(Hook *hook, int (*event)())
{
	HookEvent *hookevent;

	ASSERT(hook != NULL);

	if ((hookevent = (HookEvent *)hook_find_event(hook, event, 0)) != NULL) {
		ircdlog(LOG_ERROR, "Ignoring duplicate hook event in hook %s (already exists!)",
			hook->name);
		return NULL;
	}

	hookevent = (HookEvent *)MyMalloc(sizeof(HookEvent));
	hookevent->hook = hook;
	hookevent->event = event;
	hookevent->owner = NULL;

	dlink_add(&hook->events, hookevent);
	return hookevent;
}

void hook_del_event(HookEvent *hookevent)
{
	ASSERT(hookevent != NULL);

	dlink_del(&hookevent->hook->events, hookevent, NULL);
	MyFree(hookevent);
}

Hook *hook_add(char *name)
{
	Hook *hook;

	if ((hook = hook_find(name)) != NULL) {
		ircdlog(LOG_ERROR, "Ignoring duplicate hook %s (already exists!)", name);
		return NULL;
	}

	hook = (Hook *)MyMalloc(sizeof(Hook));
	DupString(hook->name, name);
	dlink_add(&hook_list, hook);

	return hook;
}

static void hook_del_event_from_owner(HookEvent *hookevent)
{
	Module *owner;

	ASSERT(hookevent != NULL);
	owner = hookevent->owner;
	ASSERT(owner != NULL);

	dlink_del(&owner->hookevents, hookevent, NULL);
}

void hook_del_events(Hook *hook)
{
	dlink_node *node, *next = NULL;
	HookEvent *hookevent;

	ASSERT(hook != NULL);

	DLINK_FOREACH_SAFE_DATA(hook->events.head, node, next, hookevent, HookEvent) {
		if (hookevent->owner != NULL) {
			ASSERT(hookevent->owner != hook->owner);
			sendto_realops("Hook %s contained an event from %s. Please reload "
				"this module.", hook->name, hookevent->owner->name);
			hook_del_event_from_owner(hookevent);
		}
		dlink_del(&hook->events, NULL, node);
		MyFree(hookevent);
	}
}

void hook_del(Hook *hook)
{
	ASSERT(hook != NULL);

	dlink_del(&hook_list, hook, NULL);

	if (dlink_length(&hook->events)) {
		hook_del_events(hook);
	}

	MyFree(hook->name);
	MyFree(hook);
}

void hook_report(aClient *cptr)
{
	dlink_node *node, *enode;
	Hook *hook;
	HookEvent *hookevent;

	DLINK_FOREACH_DATA(hook_list.head, node, hook, Hook) {
		send_me_notice(cptr, ":*** Hook %s: %d event(s)", hook->name, dlink_length(&hook->events));
		DLINK_FOREACH_DATA(hook->events.head, enode, hookevent, HookEvent) {
			if (hookevent->owner != NULL) {
				send_me_notice(cptr, ":*** - Hook Event from %s", hookevent->owner->name);
			}
		}
	}
}
