/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/ircd.c: core ircd functions
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: ircd.c,v 1.209 2004/05/09 20:15:04 rageircd Exp $
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "memory.h"
#include "h.h"
#include "patchlevel.h"
#include "fd.h"
#include "setup.h"
#include "isupport.h"
#include "channel.h"
#include "res.h"
#include "ssl.h"
#include "sbuf.h"
#include "conf2.h"
#include "config.h"
#include "hook.h"
#include "dlink.h"
#include "xmode.h"
#include "modules.h"
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/resource.h>
#include <errno.h>
#include <pwd.h>
#include <signal.h>
#include <fcntl.h>

static size_t initial_vmem = 0;
static int ircd_running = 0;

static char *usage[] = {
	" ",
	"RageIRCd v2.0 (bluemoon)",
	"Copyright (C) 2002 the past and present IRCd developers, and others.",
	"Refer to doc/authors/ for full credits and copyright details.",
	" ",
	"Usage: ircd [flags]",
	" ",
	"Where [flags] can be any of the following:",
	" ",
#ifdef CMDLINE_CONFIG
	"    -c N   Load N instead of default config file (ircd.conf)",
#endif
	"    -h     Displays usage information (this)",
	"    -t     Run in terminal (doesn't fork to background)",
	"    -v     Print the version to stderr and exit",
	"    -x N   Run in debug level N (log to debug.log)",
	" ",
	NULL
};

static ConfigOption server_log_files[] = {
	{ LOGFILE_DEFAULT, LOG_DEFAULT },
	{ LOGFILE_ERROR, LOG_ERROR },
	{ LOGFILE_KILL, LOG_KILL },
	{ LOGFILE_BAN, LOG_BAN },
	{ LOGFILE_CLIENT, LOG_CLIENT },
	{ LOGFILE_SERVER, LOG_SERVER },
	{ LOGFILE_OPER, LOG_OPER },
	{ LOGFILE_OVERRIDE, LOG_OVERRIDE },
	{ NULL, 0L }
};

char server_options[] = {
#ifdef DEBUGMODE
	'D',
#endif
#ifdef LOG_LIBADNS_DEBUG
	'A',
#endif
#ifdef IRCII_KLUDGE
	'K',
#endif
#ifdef STATIC_MODULES
	'S',
#endif
#ifdef USE_OPENSSL
	'E',
#endif
#ifdef USE_BLALLOC
	'B',
#endif
#ifdef USE_THROTTLE
	'T',
#endif
#ifdef USE_SELECT
	's',
#endif
#ifdef USE_POLL
	'p',
#endif
#ifdef USE_KQUEUE
	'k',
#endif
#ifdef USE_RTSIGIO
	'R',
#endif
#ifdef INCREASE_SOCK_BUFS
	'I',
#endif
#ifdef FLUD
	'F',
#endif
#ifdef OS_CYGWIN
	'W',
#endif
#ifdef CMDLINE_CONFIG
	'c',
#endif
#ifdef HAVE_MMAP
	'M',
#endif
	'\0'
};

char isupport1[512], isupport2[512], isupport3[512];
char **myargv;

aClient me, *client = &me;
struct Counter Count;
IRCd Internal;

dlink_list lunknown_list;
dlink_list lclient_list;
dlink_list lserver_list;
dlink_list oper_list;
dlink_list recvqcli_list;
dlink_list listingcli_list;

#ifdef RLIMIT_FDMAX
#define RLIMIT_FD_MAX RLIMIT_FDMAX
#else
#ifdef RLIMIT_NOFLE
#define RLIMIT_FD_MAX RLIMIT_NOFILE
#else
#ifdef RLIMIT_OPEN_MAX
#define RLIMIT_FD_MAX RLIMIT_OPEN_MAX
#else
#undef RLIMIT_FD_MAX
#endif
#endif
#endif

static size_t get_vmem()
{
	return (size_t)sbrk(0);
}

unsigned long get_max_rss()
{
	return (get_vmem() - initial_vmem);
}

static void display_usage()
{
	char **text;

	fprintf(stderr, "\nircd %s (%s)\n", ircd_version, revision);
	fprintf(stderr, "Built %s, compile # %s\n", creation, generation);

	for (text = usage; *text != NULL; text++) {
		fprintf(stderr, "%s\n", *text);
	}
}

static void save_statsfile()
{
	FILE *fd = NULL;
	if ((fd = fopen(STATS_FILE, "w+")) != NULL) {
		fprintf(fd, "%d %d %ld %ld %ld %ld %ld %ld %ld", Count.max_loc, Count.max_tot,
			Count.weekly, Count.monthly, Count.yearly, Count.start, Count.week,
			Count.month, Count.year);
		fclose(fd);
	}
}

static void init_statsfile()
{
	FILE *fd = NULL;
	char buf[1024], *item, *p;
	int cnt = 0;

	if (GeneralConfig.statsfile_save_freq) {
		add_event("save_statsfile", save_statsfile, NULL, GeneralConfig.statsfile_save_freq, 1);
	}
	if ((fd = fopen(STATS_FILE, "r")) == NULL) {
		return;
	}

	fgets(buf, 1024, fd);
	if (BadPtr(buf)) {
		fclose(fd);
		return;
	}

	for (item = strtoken(&p, buf, " "); item != NULL; item = strtoken(&p, NULL, " "), cnt++) {
		switch (cnt) {
			case 0:
				Count.max_loc = atoi(item);
				break;
			case 1:
				Count.max_tot = atoi(item);
				break;
			case 2:
				Count.weekly = atol(item);
				break;
			case 3:
				Count.monthly = atol(item);
				break;
			case 4:
				Count.yearly = atol(item);
				break;
			case 5:
				Count.start = atol(item);
				break;
			case 6:
				Count.week = atol(item);
				break;
			case 7:
				Count.month = atol(item);
				break;
			case 8:
				Count.year = atol(item);
				break;
			default:
				ircdlog(LOG_ERROR, "init_statsfile(): ignoring extra parameter (cnt %d) in %s",
					cnt, STATS_FILE);
		}
	}
	fclose(fd);

	fprintf(stderr, "Loaded IRC statistics file.\n");
}

void set_time()
{
	struct timeval new_time;

	new_time.tv_sec = 0;
	new_time.tv_usec = 0;

	if (gettimeofday(&new_time, NULL) == -1) {
		ircdlog(LOG_ERROR, "Clock failure (%d), TS may become corrupt!", errno);
		sendto_realops("Clock failure (%d), TS may become corrupt!", errno);
		restart("Clock failure!");
	}

	if (new_time.tv_sec < timeofday) {
		/* FIXME: set_back_events -- ircd-hybrid-7.0 */
	}

	Internal.ircd_time.tv_sec = new_time.tv_sec;
	Internal.ircd_time.tv_usec = new_time.tv_usec;
}

static void write_pidfile()
{
	int fd = -1, len;
	char buf[20];

	if ((fd = open(PID_FILE, O_CREAT|O_WRONLY, 0600)) == -1) {
		Debug((DEBUG_NOTICE, "Error opening pid file %s", PID_FILE));
		ircdlog(LOG_ERROR, "Error opening pid file %s", PID_FILE);
		return;
	}

	len = ircsprintf(buf, "%5d\n", (int)getpid());
	if (write(fd, buf, len) == -1) {
		Debug((DEBUG_NOTICE, "Error writing to pid file %s", PID_FILE));
		ircdlog(LOG_ERROR, "Error opening pid file %s", PID_FILE);
	}
	close(fd);
}

static int check_pid()
{
	char buf[20];
	pid_t pid = 0;
	int fd, retval = 0;

	if ((fd = open(PID_FILE, O_RDONLY)) == -1) {
		if (errno != ENOENT) {
			fprintf(stderr, "WARNING: problem opening %s: %s\n", PID_FILE, strerror(errno));
		}
		return 0;
	}

	if (read(fd, buf, sizeof(buf)) > 0) {
		if ((pid = atoi(buf)) != (int)getpid() && !kill(pid, 0)) {
			retval = (int)pid;
		}
	}
	close(fd);

	return retval;
}

static void init_internal()
{
	memset(&Internal, '\0', sizeof(Internal));

	Internal.conf_file = CONFIG_FILE;
	Internal.conf_errors = Internal.valid_conf = 0;
	Internal.debug_level = Internal.next_event = Internal.next_expire_klines = -1;
	Internal.max_con_count = Internal.max_cli_count = 1;
	Internal.curr_sendK = Internal.curr_recvK = 0;
	*Internal.masking_keys = '\0';
	Internal.default_capabs = DEFAULT_CAPABS;

#ifdef DEBUGMODE
	Internal.debug.read_calls = Internal.debug.read_errors = Internal.debug.read_errors_ignored = 0;
	Internal.debug.write_calls = Internal.debug.write_errors = Internal.debug.write_errors_ignored = 0;
#endif
}

static void init_daemon()
{
	int pid;

	if (!isatty(0)) {
		return;
	}

	if ((pid = fork()) < 0) {
		fprintf(stderr, "ERROR: couldn't fork (%s)\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	else if (pid > 0) {
		exit(EXIT_SUCCESS);
	}

	setsid();
	close(0);
}

static void init_stdio()
{
	int i, fd = -1;

	for (i = 0; i < MAXCONNECTIONS; i++) {
		if (fd_table[i].open) {
			fd_close(i);
			continue;
		}
		close(i);
	}

	for (i = 0; i < LOWEST_FD; i++) {
		fd = -1;

#ifdef DEBUGMODE
		if (Internal.debug_level >= 0 && i == 2 && !Internal.run_in_terminal) {
			truncate(LOGFILE_DEBUG, 0);
			if ((fd = open(LOGFILE_DEBUG, O_CREAT|O_WRONLY, 0600)) >= 0) {
				Debug((DEBUG_FATAL, "Debug: file %s level %d at %s", LOGFILE_DEBUG,
					Internal.debug_level, myctime(timeofday)));
				if (fd != 2) {
					dup2(fd, 2);
					close(fd);
				}
			}
		}
#endif
		if (fd < 0) {
			if ((fd = open("/dev/null", O_RDWR)) < 0) {
				exit(EXIT_FAILURE);
			}
		}
	}
}


void restart(char *message)
{
	static int been_here = 0;
	if (!been_here) {
		ircdlog(LOG_DEFAULT, "Server restarting: %s (%lu)", message, get_max_rss());
		been_here = 1;
		server_reboot();
	}
}

void server_reboot()
{
	char *errmsg;
	int i;

	sendto_ops("Aieeeeee! Restarting server... (%lu)", get_max_rss());
	Debug((DEBUG_DEBUG, "Restarting server..."));
	ircdlog(LOG_DEFAULT, "Restarting server...");

	flush_queued_buffers(me.localClient->fd);

	for (i = 3; i < MAXCONNECTIONS; i++) {
		close(i);
	}
	execv(IRCD_EXE, myargv);

	errmsg = strerror(errno);
	ircdlog(LOG_ERROR, "FATAL: execv(%s,%s) failed (%s)", IRCD_EXE, myargv[0], errmsg);
	Debug((DEBUG_FATAL, "Couldn't restart server: %s", errmsg));

	exit(EXIT_FAILURE);
}

static void parse_args(int argc, char *argv[])
{
	extern char *optarg;
	int i = -1;

	while ((i = getopt(argc, argv, "c:tvx:")) != -1) {
		switch (i) {
			case 'c':
#ifdef CMDLINE_CONFIG
				Internal.conf_file = optarg;
#else
				fprintf(stderr, "ERROR: the -c option is not enabled.\n");
				exit(EXIT_FAILURE);
#endif
				break;
			case 't':
				Internal.run_in_terminal = 1;
				break;
			case 'v':
				fprintf(stderr, "ircd %s (%s)\n", ircd_version, revision);
				fprintf(stderr, "Built %s, compile # %s\n", creation, generation);
				exit(EXIT_SUCCESS);
				break;
			case 'x':
#ifdef DEBUGMODE
				if (optarg != NULL) {
					Internal.debug_level = atoi(optarg);
					Internal.debug_mode = optarg;
				}
#else
				fprintf(stderr, "WARNING: DEBUGMODE must be defined for -x (ignoring)\n");
#endif
				break;
			default:
				display_usage();
				exit(EXIT_SUCCESS);
		}
	}
}

static void build_version()
{
	ircsprintf(ircd_version, "%s(%s)-%d.%d.%d%s%s", BASENAME, CODENAME, IRCD_VERSION,
		MAJOR_VERSION, RELEASE_VERSION, IRCD_PATCHES,
#ifdef DEBUGMODE
		"(debug)");
#else
		"");
#endif
}

void build_isupport()
{
	ircsprintf(isupport1, ISUPPORT_FEATURES1, ISUPPORT_PARAMS1);
	ircsprintf(isupport2, ISUPPORT_FEATURES2, ISUPPORT_PARAMS2);
	ircsprintf(isupport3, ISUPPORT_FEATURES3, ISUPPORT_PARAMS3);
}

static void init_core()
{
#ifdef RLIMIT_CORE
	struct rlimit ircd_lim;
	if (!getrlimit(RLIMIT_CORE, &ircd_lim)) {
		ircd_lim.rlim_cur = ircd_lim.rlim_max;
		if (setrlimit(RLIMIT_CORE, &ircd_lim)) {
			fprintf(stderr, "WARNING: unlimit core size failed (%s)\n", strerror(errno));
		}
	}
#endif
}

static void init_vmem()
{
	initial_vmem = get_vmem();
}

static void init_counter()
{
	set_time();

	Count.server = 1;
	Count.oper = 0;
	Count.chan = 0;
	Count.local = 0;
	Count.total = 0;
	Count.invisi = 0;
	Count.unknown = 0;
	Count.max_loc = 0;
	Count.max_tot = 0;
	Count.cli_restart = 0;
	Count.today = 0;
	Count.weekly = 0;
	Count.monthly = 0;
	Count.yearly = 0;
	Count.start = timeofday;
	Count.day = timeofday;
	Count.week = timeofday;
	Count.month = timeofday;
	Count.year = timeofday;
}

static void init_hash_tables()
{
	clear_client_hash_table();
	clear_channel_hash_table();
	clear_scache_hash_table();
}

static void init_maximumfds()
{
#ifdef RLIMIT_FD_MAX
	struct rlimit limit;

	if (!getrlimit(RLIMIT_FD_MAX, &limit)) {
		if (limit.rlim_max < MAXCONNECTIONS) {
			fprintf(stderr, "ERROR: ircd fd table is too big, please fix MAXCONNECTIONS!\n");
			fprintf(stderr, "       system fd limit: %ld current MAXCONNECTIONS: %d\n",
				(long)limit.rlim_max, MAXCONNECTIONS);
			exit(EXIT_FAILURE);
		}

		limit.rlim_cur = limit.rlim_max;
		if (setrlimit(RLIMIT_FD_MAX, &limit) == -1) {
			fprintf(stderr, "ERROR: failed to set max fd limit to %ld\n",
				(long)limit.rlim_max);
			exit(EXIT_FAILURE);
		}

#ifdef USE_SELECT
		if (MAXCONNECTIONS > FD_SETSIZE) {
			fprintf(stderr, "ERROR: FD_SETSIZE: %d MAXCONNECTIONS: %d\n", FD_SETSIZE,
				MAXCONNECTIONS);
			fprintf(stderr, "       Make sure your system supports a large FD_SETSIZE,\n");
			fprintf(stderr, "       then reconfigure with a max fd limit of %ld.\n",
				MAXCONNECTIONS);
			exit(EXIT_FAILURE);
		}
#endif
	}
#endif
}

static void calculate_bandwidth()
{
	static time_t last_performed = 0;
	static long last_sendK = 0, last_recvK = 0;
	long delta = (timeofday - last_performed);

	last_performed = timeofday;

	Internal.curr_sendK += ((float)(me.localClient->sendK - last_sendK) / (float)delta);
	Internal.curr_sendK /= 2;
	last_sendK = me.localClient->sendK;

	Internal.curr_recvK += ((float)(me.localClient->receiveK - last_recvK) / (float)delta);
	Internal.curr_recvK /= 2;
	last_recvK = me.localClient->receiveK;
}

static void io_loop()
{
	while (ircd_running) {
		if (Internal.next_event == -1) {
			get_next_event();
		}
		if (Internal.next_event <= timeofday) {
			run_events();
		}

		hook_run(h_pre_netio, NULL);
		engine_do_netio(1000);
		hook_run(h_post_netio, NULL);

		if (Internal.do_rehash) {
			conf_rehash();
			Internal.do_rehash = 0;
		}
	}
}

int main(int argc, char *argv[])
{
	char *line_break = "-----------------------------------------------------------------------------\n";
#ifdef USE_OPENSSL
	char *ssl_error = NULL;
#endif
	int i = 0;

	if (!geteuid()) {
		fprintf(stderr, "ERROR: never run ircd as root.\n");
		return -1;
	}

	init_core();
	init_vmem();
	init_internal();
	init_counter();
	init_sbuf();
#ifdef PROFILING
	init_profiling_signal();
#endif
	init_signal();

	build_version();

	memset((char *)&me, '\0', sizeof(me));
	myargv = argv;
	umask(077);
	parse_args(argc, argv);

	if (chdir(IRCD_PREFIX) < 0) {
		perror("chdir");
		fprintf(stderr, "ERROR: please ensure that your installation directory (IRCD_PREFIX)\n");
		fprintf(stderr, "       exists and that it is accessible.\n\n");
		fprintf(stderr, "       Current IRCD_PREFIX: %s\n", IRCD_PREFIX);
		exit(EXIT_FAILURE);
	}
	if ((i = check_pid()) > 0) {
		fprintf(stderr, "ERROR: ircd already running under process id %d.\n", i);
		exit(EXIT_FAILURE);
	}

#if !defined(SOL20) && !defined(OS_CYGWIN)
	setlinebuf(stderr);
#endif

	fd_init();

#ifdef USE_OPENSSL
	if (init_dh() == -1) {
		fprintf(stderr, "ERROR: failed to load SSL encryption layer.\n");
		exit(EXIT_FAILURE);
	}
#endif

	set_time();
	init_hash_tables();
#ifdef USE_BLALLOC
	init_blockheap(0);
#endif
	dlink_init();
#ifdef USE_BLALLOC
	init_blockheap(1);
#endif
	init_ip_hash();
	hook_init();
#ifdef PROFILING
	init_profiling_event();
#endif
	init_clients();
	init_user_mode();
	init_chan_mode();
	init_channels();
	init_channelbans();
	init_link();
	connauth_init();
#ifdef USE_THROTTLE
	init_throttle();
#endif
#ifdef FLUD
	init_fludbot();
#endif
#ifdef STATIC_MODULES
	init_whowas();
#endif
	init_command_table();
	init_modules();
	init_message_files();
	init_stats();
	init_send();

	fprintf(stderr, line_break);

	init_conf();
	check_core_commands();

	if (!Internal.conf_errors) {
#ifdef USE_OPENSSL
		if ((ssl_error = init_ssl()) != NULL) {
			report(1, "%s", ssl_error);
		}
#endif
		if (!Internal.conf_errors) {
			set_time();
			init_statsfile();
		}
	}
	if (Internal.conf_errors) {
		fprintf(stderr, "\nReceived %d fatal errors during server initialisation, terminating program.\n%s",
			Internal.conf_errors, line_break);
		exit(EXIT_FAILURE);
	}
	fprintf(stderr, "\nServer initialization successfully completed, attempting to boot.\n%s", line_break);

	set_time();
	init_maximumfds();
	build_isupport();

	strncpyzt(me.name, ServerInfo->name, sizeof(me.name));
	strncpyzt(me.info, ServerInfo->desc, sizeof(me.info));

	me.localClient = make_localclient();
	me.localClient->fd = -1;
	me.hopcount = 0;
	me.next = NULL;
	me.user = NULL;
	me.from = &me;
	SetMe(&me);
	make_server(&me);
	me.serv->up = me.name;
	me.lasttime = me.since = me.firsttime = timeofday;
	add_to_client_hash_table(me.name, &me);

	if (!init_identity(ServerInfo->identity)) {
		fprintf(stderr, "ERROR: failed to initialise identity engine.\n");
		exit(EXIT_FAILURE);
	}

	if (!Internal.run_in_terminal) {
		init_daemon();
		init_stdio();
	}

	engine_preinit();
	engine_init();

	if (!init_resolver()) {
		ircdlog(LOG_ERROR, "FATAL: init_resolver() failed to load nameserver "
			"from %s (%s)", ServerInfo->resolver_conf_file, strerror(errno));
		exit(EXIT_FAILURE);
	}

	write_pidfile();
	open_listeners();

	add_event("calculate_bandwidth", calculate_bandwidth, NULL, 2, 1);

	ircd_running = 1;
	Debug((DEBUG_NOTICE, "Server ready..."));

	set_time();
	io_loop();

	return 0;
}

void ircdlog(unsigned int log_flag, char *pattern, ...)
{
	FILE *fd;
	va_list vl;
	char buffer[1024], *current_time, *file = NULL;
	ConfigOption *opt;
	int default_log = 0;

	if (ServerInfo == NULL) {
		default_log = 1;
	}
	else if (!(ServerInfo->logs & log_flag)) {
		return;
	}

	current_time = myctime(timeofday);
	va_start(vl, pattern);
	ircvsprintf(buffer, pattern, vl);
	va_end(vl);

	if (default_log) {
		file = LOGFILE_DEFAULT;
	}
	else {
		for (opt = server_log_files; opt->token != NULL; opt++) {
			if (opt->flag == log_flag) {
				file = opt->token;
				break;
			}
		}
	}
	if (file == NULL) {
		return;
	}

	if ((fd = fopen(file, "a")) == NULL) {
		return;
	}

	fprintf(fd, "[%s] %s\n", current_time, buffer);
#ifdef DEBUGMODE
	if (log_flag == LOG_ERROR) {
		sendto_realops_lev(DEBUG_LEV, "[error log] %s", buffer);
	}
#endif

	fclose(fd);
}

#ifdef USE_ASSERT
void fatal(char *file, int line, char *statement)
{
	Debug((DEBUG_ERROR, "FATAL ERROR: false statement at %s:%d: (%s)", file, line, statement));
	ircdlog(LOG_ERROR, "FATAL ERROR: false statement at %s:%d: (%s)", file, line, statement);

	abort();
}
#endif
