/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/memory.c: internal memory management
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: memory.c,v 1.11 2004/04/10 02:18:49 rageircd Exp $
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "h.h"
#include "numeric.h"
#include "blalloc.h"
#include "setup.h"
#include "memory.h"
#include <stdlib.h>
#include <string.h>

static void mem_fail(char *file, int line)
{
#ifdef ABORT_ON_MALLOC_FAILURE
	ircdlog(LOG_ERROR, "FATAL: malloc failed at %s:%d, aborting!", file, line);
	abort();
#else
	ircdlog(LOG_ERROR, "malloc failed at %s:%d, restarting server...", file, line);
	outofmemory();
#endif
}

void *_MyMalloc(size_t size, char *file, int line)
{
	void *m;

	if ((m = calloc(1, size)) == NULL) {
		mem_fail(file, line);
	}

	return m;
}

void *_MyRealloc(void *p, size_t size, char *file, int line)
{
	void *ret;

	if (p == NULL) {
		return _MyMalloc(size, file, line);
	}
	if (!size) {
		_MyFree(p, file, line);
		return NULL;
	}
	if ((ret = realloc(p, size)) == NULL) {
		mem_fail(file, line);
	}
	return ret;
}

void _MyFree(void *p, char *file, int line)
{
	if (p != NULL) {
		free(p);
	}
}

void _DupString(char **p, const char *s, char *file, int line)
{
	*p = _MyMalloc(strlen(s) + 1, file, line);
	strcpy(*p, s);
}

void *_BlockHeapAlloc(BlockHeap *heap, char *file, int line)
{
#ifdef USE_BLALLOC
	void *ret;

	if ((ret = BlockHeapALLOC(heap)) == NULL) {
		mem_fail(file, line);
	}
	return ret;
#else
	return _MyMalloc((size_t)heap, file, line);
#endif
}

void _BlockHeapFree(BlockHeap *heap, void *p, char *file, int line)
{
#ifdef USE_BLALLOC
	ASSERT(heap != NULL);
	if (p != NULL) {
		BlockHeapFREE(heap, p);
	}
#else
	if (p != NULL) {
		_MyFree(p, file, line);
	}
#endif
}

void outofmemory(void)
{
	ircdlog(LOG_ERROR, "FATAL: out of memory, restarting server.");
	Debug((DEBUG_FATAL, "Out of memory: restarting server..."));
	restart("Out of memory");
}
