/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/module_objects.c: module object framework
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: module_objects.c,v 1.7 2004/05/08 23:23:42 rageircd Exp $
 */

#include "struct.h"
#include "setup.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "h.h"
#include "msg.h"
#include "channel.h"
#include "send.h"
#include "memory.h"
#include "patchlevel.h"
#include "hook.h"
#include "dlink.h"
#include "modules.h"
#include "xmode.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>

void clear_objects(Module *owner)
{
	dlink_node *node, *next = NULL;

	ASSERT(owner != NULL);

	DLINK_FOREACH_SAFE(owner->commands.head, node, next) {
		unregister_command_node(owner, node->data, node);
	}
	DLINK_FOREACH_SAFE(owner->events.head, node, next) {
		unregister_event_node(owner, node->data, node);
	}
	DLINK_FOREACH_SAFE(owner->hookevents.head, node, next) {
		unregister_hook_event_node(owner, node->data, node);
	}
	DLINK_FOREACH_SAFE(owner->hooks.head, node, next) {
		unregister_hook_node(owner, node->data, node);
	}
	DLINK_FOREACH_SAFE(owner->usermodes.head, node, next) {
		unregister_user_mode_node(owner, node->data, node);
	}
}

Command *register_command(Module *owner, msg_ptr *msg, int (*func)())
{
	Command *cmd;

	ASSERT(owner != NULL);
	ASSERT(msg != NULL);

	if ((cmd = add_command(msg, func)) == NULL) {
		sendto_realops("Couldn't register command %s in module %s", msg->msg_str,
			owner->name);
		return NULL;
	}
	cmd->owner = owner;

	dlink_add(&owner->commands, cmd);
	return cmd;
}

Event *register_event(Module *owner, char *name, void (*func)(), void *data, time_t when, int loop)
{
	Event *event;

	ASSERT(owner != NULL);
	ASSERT(!BadPtr(name));

	if ((event = add_event(name, func, data, when, loop)) == NULL) {
		sendto_realops("Couldn't register event %s in module %s", name, owner->name);
		return NULL;
	}
	event->owner = owner;

	dlink_add(&owner->events, event);
	return event;
}

HookEvent *register_hook_event(Module *owner, Hook *hook, int (*event)())
{
	HookEvent *hookevent;

	ASSERT(owner != NULL);
	ASSERT(hook != NULL);

	if ((hookevent = hook_add_event(hook, event)) == NULL) {
		sendto_realops("Couldn't register event for hook %s in module %s", hook->name,
			owner->name);
		return NULL;
	}
	hookevent->owner = owner;

	dlink_add(&owner->hookevents, hookevent);
	return hookevent;
}

Hook *register_hook(Module *owner, char *name)
{
	Hook *hook;

	ASSERT(owner != NULL);
	ASSERT(!BadPtr(name));

	if ((hook = hook_add(name)) == NULL) {
		sendto_realops("Couldn't register hook %s in module %s", name, owner->name);
		return NULL;
	}
	hook->owner = owner;

	dlink_add(&owner->hooks, hook);
	return hook;
};

xMode *register_user_mode(Module *owner, char flag, long *mode, unsigned int options)
{
	xMode *t;

	ASSERT(owner != NULL);

	if ((t = add_user_mode(flag, mode, options)) == NULL) {
		return NULL;
	}
	t->owner = owner;

	dlink_add(&owner->usermodes, t);
	return t;
}

int unregister_command_node(Module *owner, Command *cmd, dlink_node *node)
{
	ASSERT(owner != NULL);
	ASSERT(cmd != NULL);

	dlink_del(&owner->commands, cmd, node);

	if (!del_command_cmd(cmd)) {
		sendto_realops("Couldn't unregister command %s from module %s",
			cmd->msg->msg_str, owner->name);
		return 0;
	}
	return 1;
}

int unregister_event_node(Module *owner, Event *event, dlink_node *node)
{
	ASSERT(owner != NULL);
	ASSERT(event != NULL);

	dlink_del(&owner->events, event, node);
	del_event(event, NULL);
	return 1;
}

int unregister_hook_event_node(Module *owner, HookEvent *hookevent, dlink_node *node)
{
	ASSERT(owner != NULL);
	ASSERT(hookevent != NULL);

	dlink_del(&owner->hookevents, hookevent, node);
	hook_del_event(hookevent);
	return 1;
}

int unregister_hook_node(Module *owner, Hook *hook, dlink_node *node)
{
	ASSERT(owner != NULL);
	ASSERT(hook != NULL);

	dlink_del(&owner->hooks, hook, node);
	hook_del(hook);
	return 1;
}

int unregister_user_mode_node(Module *owner, xMode *t, dlink_node *node)
{
	ASSERT(owner != NULL);
	ASSERT(t != NULL);

	dlink_del(&owner->usermodes, t, node);
	del_user_mode(t->flag, &t->mode);
	return 1;
}
