/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/modules.c: dynamic module routenes
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: modules.c,v 1.98 2004/05/13 17:43:52 rageircd Exp $
 */

#include "struct.h"
#include "setup.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "h.h"
#include "msg.h"
#include "channel.h"
#include "send.h"
#include "memory.h"
#include "patchlevel.h"
#include "hook.h"
#include "dlink.h"
#include "modules.h"
#include "xmode.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <limits.h>

void init_modules()
{
#ifndef STATIC_MODULES
	add_command(&CMD_MODULE, m_module);
#else
	commands_load();
	aliases_load();
	antidrone_load();
#endif
}

#ifndef STATIC_MODULES

#include <dlfcn.h>
#include <dirent.h>

static const char *mod_err_tab[] = {
	"object not loaded",
	"object already loaded",
	"missing module header",
	"invalid module header",
	"interface version too old",
	"missing module load",
	"missing module unload",
	"cannot unload permanent modules",
	"module load failure",
	"module unload failure",
	NULL
};

static int load_one_module(char *, Module **);
static int unload_one_module(char *, int);
static int reload_one_module(char *, Module **);

#ifndef RTLD_NOW
#define RTLD_NOW RTLD_LAZY
#endif

dlink_list module_list;

int do_module(char *call, char *name, aClient *sptr)
{
	int retval = 0;
	Module *header = NULL;
	const char *mod_err;

	ASSERT(!BadPtr(call));
	ASSERT(!BadPtr(name));

	if (!mycmp(call, "LOAD")) {
		retval = load_one_module(name, &header);
		call = "load";
	}
	else if (!mycmp(call, "UNLOAD")) {
		retval = unload_one_module(name, 0);
		call = "unload";
	}
	else if (!mycmp(call, "RELOAD")) {
		retval = reload_one_module(name, &header);
		call = "reload";
	}
	else {
		return MOD_UNKNOWN;
	}

	if (retval == MOD_SUCCESS) {
		if (sptr != NULL) {
			sendto_realops("Successfully %sed %s [%s] by request of %s", call,
				(header != NULL) ? header->name : name,
				(header != NULL) ? header->revision : "N/A",
				sptr->name);
		}
		return MOD_SUCCESS;
	}

	if ((mod_err = dlerror()) != NULL) {
		if (sptr != NULL) {
			send_me_notice(sptr, ":Couldn't %s %s", call, mod_err);
		}
		else {
			report(0, "%s", mod_err);
		}
	}
	else {
		if (sptr != NULL) {
			send_me_notice(sptr, ":Couldn't %s %s: %s", call, name, mod_err_tab[retval]);
		}
		else if (retval > MOD_ERR_ALREADY_LOADED) {
			report(0, "%s: failed to load (%s)", name, mod_err_tab[retval]);
		}
	}

	return MOD_FAILURE;
}

int do_module_list(aClient *sptr)
{
	unsigned long length = dlink_length(&module_list);
	dlink_node *node;
	Module *header;

	send_me_notice(sptr, ":There %s currently %d module%s loaded", (length == 1) ? "is" : "are",
		length, (length == 1) ? "" : "s");
	send_me_notice(sptr, ":The current module version is %d (minimum %d)", CUR_MODULE_VERSION,
		MIN_MODULE_VERSION);

	if (length) {
		send_me_noticeNA(sptr, ":--------------------------------------------------------------");
		DLINK_FOREACH_DATA(module_list.head, node, header, Module) {
			send_me_notice(sptr, ":%s: %s [%s] (%d)", header->name,
				!BadPtr(header->info) ? header->info : "unknown",
				!BadPtr(header->revision) ? header->revision : "N/A",
				header->version);
		}
		send_me_noticeNA(sptr, ":--------------------------------------------------------------");
	}

	return 0;
}

static dlink_node *find_module_node_byname(char *name)
{
	dlink_node *node;
	Module *header;

	DLINK_FOREACH_DATA(module_list.head, node, header, Module) {
		if (!irccmp(header->name, name)) {
			return node;
		}
	}

	return NULL;
}

static void *get_symbol(char *symbol, void *source_ptr)
{
	char _symbol[512];
	void *symbol_ptr = NULL;
	const char *errstr = NULL;
	dlink_node *node;
	Module *header;

	ircsprintf(_symbol, "_%s", symbol);

	if (source_ptr != NULL) {
		symbol_ptr = dlsym(source_ptr, symbol);
		if ((errstr = dlerror()) == NULL) {
			return symbol_ptr;
		}

		symbol_ptr = dlsym(source_ptr, _symbol);
		if ((errstr = dlerror()) == NULL) {
			return symbol_ptr;
		}

		return NULL;
	}

	DLINK_FOREACH_DATA(module_list.head, node, header, Module) {
		symbol_ptr = dlsym(header->ptr, symbol);
		if ((errstr = dlerror()) == NULL) {
			return symbol_ptr;
		}

		symbol_ptr = dlsym(header->ptr, _symbol);
		if ((errstr = dlerror()) == NULL) {
			return symbol_ptr;
		}
	}

	return NULL;
}

void *get_symbol_from_source(char *name, char *source)
{
	void *symbol;

	if ((symbol = get_symbol(name, NULL)) != NULL) {
		return symbol;
	}

	if (do_module("load", source, NULL) == MOD_SUCCESS) {
		if ((symbol = get_symbol(name, NULL)) != NULL) {
			return symbol;
		}
	}

	return NULL;
}

static int load_one_module(char *name, Module **header_ptr)
{
	void *ptr;
	int (*loadfunc)(), (*unloadfunc)();
	Module *header;

	if (find_module_node_byname(name) != NULL) {
		return MOD_ERR_ALREADY_LOADED;
	}

	if ((ptr = dlopen(name, RTLD_NOW)) == NULL) {
		return MOD_FAILURE;
	}

	if ((header = get_symbol("module_header", ptr)) == NULL) {
		dlclose(ptr);
		return MOD_ERR_HEADER_MISSING;
	}

	if (BadPtr(header->name)) {
		dlclose(ptr);
		return MOD_ERR_HEADER_INVALID;
	}

	if (header->version < MIN_MODULE_VERSION) {
		dlclose(ptr);
		return MOD_ERR_OLD_VERSION;
	}
	else if (header->version != CUR_MODULE_VERSION) {
		sendto_realops_lev(DEBUG_LEV, "%s module interface is version %d, yet the servers "
			"current version is %d. It should be compatible, but you are recommended "
			"to update.", header->name, header->version, CUR_MODULE_VERSION);
	}

	if ((loadfunc = get_symbol("module_load", ptr)) == NULL) {
		dlclose(ptr);
		return MOD_ERR_LOAD_MISSING;
	}

	if ((unloadfunc = get_symbol("module_unload", ptr)) == NULL) {
		dlclose(ptr);
		return MOD_ERR_UNLOAD_MISSING;
	}

	if ((*loadfunc)() != MOD_SUCCESS) {
		dlclose(ptr);
		return MOD_ERR_LOAD_FAILED;
	}

	header->ptr = ptr;
	header->unloadfunc = unloadfunc;

	if (header_ptr != NULL) {
		*header_ptr = header;
	}

	dlink_add(&module_list, header);

	return MOD_SUCCESS;
}

static inline int do_module_unload(Module *header, dlink_node *node)
{
	ASSERT(header != NULL);
	ASSERT(node != NULL);

	if ((*header->unloadfunc)() == MOD_FAILURE) {
		return MOD_FAILURE;
	}

	clear_objects(header);
	dlclose(header->ptr);
	dlink_del(&module_list, NULL, node);

	return MOD_SUCCESS;
}

static int unload_one_module(char *name, int reloading)
{
	dlink_node *node;
	Module *header;

	if ((node = find_module_node_byname(name)) == NULL) {
		return MOD_ERR_NOT_LOADED;
	}
	header = (Module *)node->data;

	if (!reloading && (header->flags & MOD_FLAG_PERM)) {
		return MOD_ERR_PERM;
	}

	if (do_module_unload(header, node) != MOD_SUCCESS) {
		return MOD_ERR_UNLOAD_FAILED;
	}

	return MOD_SUCCESS;
}

void unload_all_modules()
{
	dlink_node *node, *next = NULL;
	Module *header;

	DLINK_FOREACH_SAFE_DATA(module_list.head, node, next, header, Module) {
		if (do_module_unload(header, node) == MOD_FAILURE) {
			ircdlog(LOG_ERROR, "unload_all_modules() encountered MOD_FAILURE with %s", header->name);
			sendto_realops("Warning! %s failed to unload during server rehash.", header->name);
		}
	}
}

static int reload_one_module(char *name, Module **header)
{
	int i;

	if ((i = unload_one_module(name, 1)) != MOD_SUCCESS) {
		return i;
	}

	return load_one_module(name, header);
}

void load_modules_from_path(char *mod_path)
{
	DIR *mod_dir;
	struct dirent *file_ptr;
	char mod_name[PATH_MAX + 1];
	int len;

	if ((mod_dir = opendir(mod_path)) == NULL) {
		char *error_msg = strerror(errno);
		ircdlog(LOG_ERROR, "Couldn't load modules from path %s: %s", mod_path,
			error_msg);
		report(1, "ERROR: couldn't load modules from path %s: %s", mod_path,
			error_msg);
		return;
	}

	while ((file_ptr = readdir(mod_dir)) != NULL) {
		if ((len = strlen(file_ptr->d_name)) <= 3) {
			continue;
		}
		if ((file_ptr->d_name[len - 3] == '.') && (file_ptr->d_name[len - 2] == 's')
		  && (file_ptr->d_name[len - 1] == 'o')) {
			ircsprintf(mod_name, "%s/%s", mod_path, file_ptr->d_name);
			do_module("load", mod_name, NULL);
		}
	}

	closedir(mod_dir);
}

/*
 * m_module
 *	parv[0] = sender prefix
 *	parv[1] = function
 *	parv[2] = module
 */
int m_module(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	if (parc < 2 || *parv[1] == '\0') {
		send_me_numeric(sptr, ERR_NEEDMOREPARAMS, "MODULE");
		return 0;
	}

	if (!mycmp(parv[1], "LIST")) {
		return do_module_list(sptr);
	}

	if (!HasMode(sptr, UMODE_NETADMIN|UMODE_ADMIN)) {
		send_me_numericNA(sptr, ERR_NOPRIVILEGES);
		return 0;
	}

	if (parc < 3 || *parv[2] == '\0') {
		send_me_numeric(sptr, ERR_NEEDMOREPARAMS, "MODULE");
		return 0;
	}

	if (do_module(parv[1], parv[2], sptr) == MOD_UNKNOWN) {
		send_me_noticeNA(sptr, ":Syntax: /MODULE <load|unload|reload|list> [module.so]");
	}

	return 0;
}

#endif
