/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/s_serv.c: server functions
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: s_serv.c,v 1.136 2004/04/09 01:45:53 rageircd Exp $
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "patchlevel.h"
#include "ssl.h"
#include "zlink.h"
#include "memory.h"
#include "xmode.h"
#include <sys/stat.h>
#include <fcntl.h>
#include <utmp.h>
#include "h.h"
#if defined(HAVE_STRING_H)
#include <string.h>
#else
#include <strings.h>
#endif
#if defined(AIX) || defined(SVR3)
#include <time.h>
#endif

Capability cap_table[] = {
	/* CAP_FLAG,	TOKEN,		link, synch, PrettyToken */
	{ CAP_BURST,	"BURST",	0, 0, NULL },
	{ CAP_UNCONN,	"UNCONNECT",	0, 0, NULL },
	{ CAP_ULINE,	NULL,		1, 0, "U-Lined" },
	{ CAP_ZIP,	"ZIP",		1, 0, "OutputZipped" },
#ifdef USE_OPENSSL
	{ CAP_DKEY,	"DKEY",		1, 0, "Encrypted" },
#endif
	{ CAP_SSJ3,	"SSJ3",		0, 1, NULL },
	{ CAP_SN2,	"SN2",		0, 1, NULL },
	{ CAP_VHOST,	"VHOST",	0, 1, NULL },
	{ CAP_SUID,	"SUID",		1, 0, "Identities" },
	{ CAP_TOK1,	"TOK1",		1, 0, "Tokenised" },
	{ 0, NULL, 0, 0, NULL }
};

char *make_capab_list(unsigned int capabs, int pretty, int type)
{
	static char cbuf[512];
	int cidx = 0;
	Capability *c;

	cbuf[cidx] = '\0';
	for (c = cap_table; c->flag; c++) {
		if (!(capabs & c->flag) || (type < 0 && !c->link_state) || (type > 0 && !c->synch_state)) {
			continue;
		}
		if (pretty && (c->pretty_token != NULL)) {
			ADD_STRING(c->pretty_token, cbuf, cidx);
		}
		else if (c->token != NULL) {
			ADD_STRING(c->token, cbuf, cidx);
		}
	}
	return cbuf;
}

void try_connections(void *unused)
{
	dlink_node *node;
	ConfigItem_link *linkp;
	ConfigItem_ban *ban;
	aClient *cptr;

	Debug((DEBUG_DEBUG, "Connection check at: %s", myctime(timeofday)));

	DLINK_FOREACH_DATA(conf_link_list.head, node, linkp, ConfigItem_link) {
		if (!(linkp->flags & LINK_AUTOCONNECT) || (linkp->next_connect > timeofday)) {
			continue;
		}

		linkp->next_connect = (timeofday + GeneralConfig.auto_connect_freq);

		if ((cptr = find_server(linkp->servername, NULL)) != NULL) {
			continue;
		}
		if (linkp->class->clients > linkp->class->max_clients) {
			continue;
		}
		if ((ban = find_ban(BAN_SERVER, linkp->servername)) != NULL) {
			ircdlog(LOG_BAN, "Ignoring auto-connect to %s (banned: %s)",
				linkp->servername,
				BadPtr(ban->reason) ? "no reason" : ban->reason);
			continue;
		}

		if (serv_connect(linkp, NULL)) {
			sendto_gnotice("Connection to %s[%s].%d activated.", linkp->servername,
				linkp->host, linkp->port);
		}
	}

	Debug((DEBUG_DEBUG, "Next connection check at: %s", myctime(timeofday + GeneralConfig.auto_connect_freq)));
	return;
}

static inline void sendnick_TS_no_token1(aClient *sptr, aClient *acptr, char *ubuf, char *maskbuf)
{
	if (CapSN2(sptr)) {
		sendto_one_client_nopostfix(sptr, NULL, &CMD_SNICK, "%s %ld %d %s %s %lu %s %s %lu %s :%s",
			acptr->name, acptr->tsinfo, acptr->hopcount + 1, acptr->username,
			acptr->host, htonl(acptr->ip.s_addr), maskbuf, acptr->user->server, 
			acptr->user->servicestamp, ubuf, acptr->info);
		return;
	}

	sendto_one_client_nopostfix(sptr, NULL, &CMD_NICK, "%s %d %ld %s %s %s %s %lu %lu :%s", acptr->name,
		acptr->hopcount + 1, acptr->tsinfo, ubuf, acptr->username, acptr->host,
		acptr->user->server, acptr->user->servicestamp, htonl(acptr->ip.s_addr), acptr->info);

	if (CapVHOST(sptr) && (*maskbuf != '*')) {
		sendto_one_client_nopostfix(sptr, NULL, &CMD_VHOST, "%s :%s", acptr->name, maskbuf);
	}

	return;
}

static inline void sendnick_TS_has_identity(aClient *sptr, aClient *acptr, char *ubuf, char *maskbuf)
{
	if (!CapID(sptr)) {
		sendnick_TS_no_token1(sptr, acptr, ubuf, maskbuf);
		return;
	}

	if (CapSN2(sptr)) {
		sendto_one_client_nopostfix(sptr, NULL, &CMD_SNICK, "%s %B %d %s %s %B %s !%s %B %s :%s",
			acptr->name, acptr->tsinfo, acptr->hopcount + 1, acptr->username,
			acptr->host, acptr->ip.s_addr, maskbuf, acptr->id.string,
			acptr->user->servicestamp, ubuf, acptr->info);
		return;
	}

	sendto_one_client_nopostfix(sptr, NULL, &CMD_NICK, "%s %d %B %s %s %s !%s %B %B :%s", acptr->name,
		acptr->name, acptr->hopcount + 1, acptr->tsinfo, ubuf, acptr->username, acptr->host,
		acptr->id.string, acptr->user->servicestamp, acptr->ip.s_addr, acptr->info);

	if (CapVHOST(sptr)) {
		sendto_one_client_nopostfix(sptr, NULL, &CMD_VHOST, "%s :%s", acptr->name, maskbuf);
	}

	return;
}

static inline void sendnick_TS_no_identity(aClient *sptr, aClient *acptr, char *ubuf, char *maskbuf)
{
	if (!CapID(sptr)) {
		sendnick_TS_no_token1(sptr, acptr, ubuf, maskbuf);
		return;
	}

	if (CapSN2(sptr)) {
		sendto_one_client_nopostfix(sptr, NULL, &CMD_SNICK, "%s %B %d %s %s %B %s %s %B %s :%s",
			acptr->name, acptr->tsinfo, acptr->hopcount + 1, acptr->username,
			acptr->host, acptr->ip.s_addr, maskbuf, acptr->user->server,
			acptr->user->servicestamp, ubuf, acptr->info);
		return;
	}

	sendto_one_client_nopostfix(sptr, NULL, &CMD_NICK, "%s %d %B %s %s %s %s %B %B :%s",
		acptr->name, acptr->hopcount + 1, acptr->tsinfo, ubuf, acptr->username,
		acptr->host, acptr->user->server, acptr->user->servicestamp, acptr->ip.s_addr,
		acptr->info);

	if (CapVHOST(sptr) && (*maskbuf != '*')) {
		sendto_one_client_nopostfix(sptr, NULL, &CMD_VHOST, "%s :%s", acptr->name, maskbuf);
	}

	return;
}

static void sendnick_TS(aClient *sptr, aClient *acptr)
{
	char ubuf[16], *maskbuf;

	if (!IsPerson(acptr)) {
		return;
	}

	send_umode(NULL, acptr, 0, SEND_UMODES, ubuf);
	if (*ubuf == '\0') {
		ubuf[0] = '+';
		ubuf[1] = '\0';
	}

	maskbuf = (HasMode(acptr, UMODE_MASKED) ? acptr->user->maskedhost : "*");
	if (HasSUID(acptr)) {
		sendnick_TS_has_identity(sptr, acptr, ubuf, maskbuf);
	}
	else {
		sendnick_TS_no_identity(sptr, acptr, ubuf, maskbuf);
	}
	return;
}

int do_server_estab(aClient *cptr)
{
	char *inpath, *cap_list;
	aClient *acptr;
	aChannel *chptr;
	ConfigItem_ban *ban;
	dlink_node *node = NULL;
	unsigned long link_capabs;

	ASSERT(cptr != NULL);
	ASSERT(cptr->localClient != NULL);

	inpath = get_client_name(cptr, HIDE_IP);

	SetServer(cptr);
	Count.server++;
	Count.myserver++;

	if (CapZIP(cptr) && WantZIP(cptr)) {
		sendto_one_client_nopostfix(cptr, NULL, &CMD_SVINFO, "ZIP");
		SetOutputZIP(cptr);
		cptr->serv->zip_out = zip_create_output_session();
	}

#ifdef INCREASE_SOCK_BUFS
	increase_sock_bufs(cptr->localClient->fd, 1);
#endif

	dlink_del_nofree(&lunknown_list, NULL, &cptr->localClient->self);
	dlink_add_node(&lserver_list, &cptr->localClient->self, cptr);
	Count.unknown--;

	attach_class(cptr);
	if ((acptr = find_server(cptr->name, NULL)) != NULL) {
		aClient *bcptr = (cptr->firsttime > acptr->from->firsttime) ? cptr : acptr->from;
		char *b_inpath = get_client_name(bcptr, HIDE_IP);

		sendto_one_client_nopostfix(bcptr, NULL, &CMD_ERROR, ":Server %s already exists", cptr->name);
		if (bcptr == cptr) {
			sendto_gnotice("Link %s cancelled, server %s already exists (final phase)",
				b_inpath, cptr->name);
			sendto_serv_msg_butone(bcptr, &me, NULL, &CMD_GNOTICE, ":Link %s cancelled, "
				"server %s already exists (final phase)", b_inpath, cptr->name);
			return exit_client(bcptr, bcptr, &me, "Server Exists (final phase)");
		}

		sendto_gnotice("Link %s cancelled, server %s reintroduced by %s (final phase)",
			b_inpath, cptr->name, inpath);
		sendto_serv_msg_butone(bcptr, &me, NULL, &CMD_GNOTICE, ":Link %s cancelled, server %s "
			"reintroduced by %s (final phase)", b_inpath, cptr->name, inpath);
		return exit_client(bcptr, bcptr, &me, "Server Exists (final phase)");
	}
	if (*cptr->id.string != '\0') {
		char b64id[8];

		strcpy(b64id, cptr->id.string);

		if (!add_base64_serv(cptr, b64id)) {
			char *name = "<UNKNOWN>";

			if ((acptr = find_serv_by_base64_id(b64id, NULL)) != NULL) {
				name = acptr->name;
			}

			sendto_gnotice("Link %s cancelled, identity collision with %s (final phase)",
				inpath, name);
			sendto_serv_msg_butone(cptr, &me, NULL, &CMD_GNOTICE, ":Link %s cancelled, "
				"identity collision with %s (final phase)", inpath, name);
			return exit_client(cptr, cptr, &me, "Identity collision (final phase)");
		}
	}

	if (find_uline(cptr->name)) {
		Count.myulined++;
		SetULine(cptr);
	}

	link_capabs = cptr->localClient->capabs;
	if (!OutputZIP(cptr)) {
		link_capabs &= ~CAP_ZIP;
	}
#ifdef USE_OPENSSL
	if (!InputRC4(cptr) || !OutputRC4(cptr)) {
		link_capabs &= ~CAP_DKEY;
	}
#endif
	if (IsULine(cptr)) {
		link_capabs |= CAP_ULINE;
	}

	cap_list = get_link_cap(link_capabs);
	sendto_gnotice("Link with %s established (capabilities: %s)", inpath,
		!BadPtr(cap_list) ? cap_list : "none");
	sendto_serv_msg_butone(NULL, &me, NULL, &CMD_GNOTICE, ":Link with %s established "
		"(capabilities: %s)", inpath, !BadPtr(cap_list) ? cap_list : NULL);
	ircdlog(LOG_SERVER, "link with %s established (%s)", get_client_name(cptr, FALSE),
		cap_list);

	add_to_client_hash_table(cptr->name, cptr);
	find_or_add(cptr->name);

	sendto_serv_capab_msg_butone(cptr, &me, NULL, 0, CAP_ID, &CMD_SERVER,
		"%s 2 :%s", cptr->name, cptr->info);
	if (HasSUID(cptr)) {
		sendto_serv_capab_msg_butone(cptr, &me, NULL, CAP_ID, 0, &CMD_SERVER,
			"%s 2 !%s :%s", cptr->name, cptr->id.string, cptr->info);
	}
	else {
		sendto_serv_capab_msg_butone(cptr, &me, NULL, CAP_ID, 0, &CMD_SERVER,
			"%s 2 :%s", cptr->name, cptr->info);
	}

	for (acptr = &me; acptr; acptr = acptr->prev) {
		if (acptr->from == cptr || !IsServer(acptr)) {
			continue;
		}

		if (!CapID(cptr) || (CapID(cptr) && (!HasSUID(acptr->uplink)
		  || (HasSUID(acptr->uplink) && !HasSUID(acptr))))) {
			sendto_one_client_nopostfix(cptr, acptr->uplink, &CMD_SERVER, "%s %d :%s",
				acptr->name, acptr->hopcount + 1, acptr->info);
		}
		else {
			sendto_one_client_nopostfix(cptr, acptr->uplink, &CMD_SERVER, "%s %d !%s :%s",
				acptr->name, acptr->hopcount + 1, acptr->id.string, acptr->info);
		}
	}

	DLINK_FOREACH_DATA(conf_ban_list.head, node, ban, ConfigItem_ban) {
		if (!ServerInfo->hub) {
			dlink_del(&conf_ban_list, NULL, node);
			free_ban(ban);
			continue;
		}

		if ((ban->flags & (BAN_NICKNAME|BAN_SVS)) == (BAN_NICKNAME|BAN_SVS)) {
			sendto_one_client_nopostfix(cptr, &me, &CMD_SQLINE, "%s :%s", ban->mask,
				BadPtr(ban->reason) ? "reserved" : ban->reason);
		}
		else if ((ban->flags & (BAN_GCOS|BAN_SVS)) == (BAN_GCOS|BAN_SVS)) {
			sendto_one_client_nopostfix(cptr, &me, &CMD_SGLINE, "%d :%s:%s",
				strlen(ban->mask), ban->mask,
				BadPtr(ban->reason) ? "reserved" : ban->reason);
		}
	}

	if (CapBURST(cptr)) {
		sendto_one_client_nopostfix(cptr, NULL, &CMD_BURST, "");
	}

	{
		chanMember *cm;
		static char nickissent = 1;

		synch_chan_modes(cptr, NULL);

		nickissent = 3 - nickissent;
		for (chptr = channel; chptr != NULL; chptr = chptr->nextch) {
			for (cm = chptr->members; cm != NULL; cm = cm->nextuser) {
				acptr = cm->cptr;
				if (!IsPerson(acptr)) {
					continue;
				}
				if (acptr->user->nicksent != nickissent) {
					acptr->user->nicksent = nickissent;
					if (acptr->from != cptr) {
						sendnick_TS(cptr, acptr);
					}
				}
			}
			synch_chan_modes(cptr, chptr);
		}
		for (acptr = &me; acptr != NULL; acptr = acptr->prev) {
			if (!IsPerson(acptr)) {
				continue;
			}
			if (acptr->user->nicksent != nickissent) {
				acptr->user->nicksent = nickissent;
				if (acptr->from != cptr) {
					sendnick_TS(cptr, acptr);
				}
			}
		}
	}

	if (OutputZIP(cptr)) {
		unsigned long in_b, out_b;
		double ratio;

		zip_out_get_stats(cptr->serv->zip_out, &in_b, &out_b, &ratio);
		if (in_b) {
			sendto_gnotice("Connect burst to %s: %lu bytes normal, %lu compressed (%3.2f%%)",
				inpath, in_b, out_b, ratio);
			sendto_serv_msg_butone(cptr, &me, NULL, &CMD_GNOTICE, ":Connect burst to %s: %lu "
				"bytes normal, %lu compressed (%3.2f%%)", inpath, in_b, out_b, ratio);
		}
	}

	SetPingSent(cptr);
	SetUserBurst(cptr);
	SetTopicBurst(cptr);

	if (CapBURST(cptr)) {
		SetSentSOB(cptr);
	}

	sendto_one_client_nopostfix(cptr, &me, &CMD_PING, ":%s", me.name);
	return 0;
}

int send_lusers(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	static long last_time = 0;
	static int s_count = 0, c_count = 0, u_count = 0, i_count = 0;
	static int o_count = 0, my_client = 0, my_server = 0, my_ulined = 0;
	static int ch_count = 0;
	int forced = (HasMode(sptr, UMODE_OPER) && parc > 3);
	aClient *acptr;
	aChannel *chptr;

	my_server = Count.myserver;
	my_ulined = Count.myulined;
	my_client = Count.local;
	i_count = Count.invisi;
	u_count = Count.unknown;
	c_count = Count.total - Count.invisi;
	s_count = Count.server;
	o_count = Count.oper;
	ch_count = Count.chan;

	if (forced || last_time + LUSERS_CACHE_TIME > timeofday) {
		last_time = timeofday;
		s_count = c_count = u_count = i_count = o_count = my_client = my_server = my_ulined = 0;

		for (acptr = client; acptr; acptr = acptr->next) {
			switch (acptr->status) {
				case STAT_SERVER:
					if (MyConnect(acptr)) {
						my_server++;
						if (IsULine(acptr)) {
							my_ulined++;
						}
					}
				case STAT_ME:
					s_count++;
					break;
				case STAT_CLIENT:
					if (HasMode(acptr, UMODE_OPER)) {
						o_count++;
					}
					if (MyConnect(acptr)) {
						if (GeneralConfig.show_invisible_lusers
						  || (!GeneralConfig.show_invisible_lusers
						  && HasMode(acptr, UMODE_INVISIBLE)
						  && HasMode(sptr, UMODE_OPER))) {
							my_client++;
						}
					}
					if (!HasMode(acptr, UMODE_INVISIBLE)) {
						c_count++;
					}
					else {
						i_count++;
					}
					break;
				default:
					u_count++;
					break;
			}
		}
		for (chptr = channel; chptr != NULL; chptr = chptr->nextch) {
			ch_count++;
		}
		if (!forced) {
			if (my_server != Count.myserver) {
				sendto_realops_lev(DEBUG_LEV, "Local server count off by %d",
					Count.myserver - my_server);
				Count.myserver = my_server;
			}
			if (my_ulined != Count.myulined) {
				sendto_realops_lev(DEBUG_LEV, "Local ulined server count off by %d",
					Count.myulined - my_ulined);
				Count.myulined = my_ulined;
			}
			if (my_client != Count.local) {
				sendto_realops_lev(DEBUG_LEV, "Local client count off by %d",
					Count.local - my_client);
				Count.local = my_client;
			}
			if (s_count != Count.server) {
				sendto_realops_lev(DEBUG_LEV, "Server count off by %d",
					Count.server - s_count);
				Count.server = s_count;
			}
			if (i_count != Count.invisi) {
				sendto_realops_lev(DEBUG_LEV, "Invisible client count off by %d",
					Count.invisi - i_count);
				Count.invisi = i_count;
			}
			if (c_count + i_count != Count.total) {
				sendto_realops_lev(DEBUG_LEV, "Total client count off by %d",
					Count.total - (c_count + i_count));
				Count.total = c_count + i_count;
			}
			if (o_count != Count.oper) {
				sendto_realops_lev(DEBUG_LEV, "Oper count off by %d",
					Count.oper - o_count);
				Count.oper = o_count;
			}
			if (ch_count != Count.chan) {
				sendto_realops_lev(DEBUG_LEV, "Chan count off by %d",
					Count.chan - ch_count);
				Count.chan = ch_count;
			}
			Count.unknown = u_count;
		}
	}

	if (GeneralConfig.show_invisible_lusers || (!GeneralConfig.show_invisible_lusers
	  && HasMode(sptr, UMODE_OPER) && (i_count > 0))) {
		send_me_numeric(sptr, RPL_LUSERCLIENT, c_count, i_count, s_count);
	}
	else {
		send_me_numeric_buf(sptr, ":There are %d users on %d servers", RPL_LUSERCLIENT,
			c_count, s_count);
	}
	if (o_count > 0) {
		send_me_numeric(sptr, RPL_LUSEROP, o_count);
	}
	if (u_count > 0) {
		send_me_numeric(sptr, RPL_LUSERUNKNOWN, u_count);
	}
	if (ch_count > 0) {
		send_me_numeric(sptr, RPL_LUSERCHANNELS, ch_count);
	}
	send_me_numeric(sptr, RPL_LUSERME, my_client,
		(GeneralConfig.hide_ulined_servers && !HasMode(sptr, UMODE_OPER)) ? my_server - my_ulined : my_server);

	send_me_numeric(sptr, RPL_LOCALUSERS, Count.local, Count.max_loc);
	send_me_numeric(sptr, RPL_GLOBALUSERS, Count.total, Count.max_tot);
	send_me_numeric(sptr, RPL_STATSCONN, Internal.max_con_count, Internal.max_cli_count, Count.cli_restart);

	if (my_client > Internal.max_cli_count) {
		Internal.max_cli_count = my_client;
	}
	if (my_client + my_server > Internal.max_con_count) {
		Internal.max_con_count = my_client + my_server;
	}
	return 0;
}
