/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/s_user.c: user functions
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: s_user.c,v 1.192 2004/05/12 01:51:18 rageircd Exp $
 */

#include "struct.h"
#include "patchlevel.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include "ssl.h"
#include "isupport.h"
#include "hook.h"
#include "dlink.h"
#include "xmode.h"
#include <sys/stat.h>
#include <utmp.h>
#include <fcntl.h>
#ifdef HAVE_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif

BlockHeap *client_heap;
BlockHeap *localclient_heap;

BlockHeap *user_heap;
BlockHeap *localuser_heap;

static int check_local_bans(aClient *cptr)
{
	dlink_node *node;
	ConfigItem_ban *ban;
	char *userhost;

	if (!MyClient(cptr)) {
		return 0;
	}

	userhost = make_user_host(cptr->username, cptr->host);
	DLINK_FOREACH_DATA(conf_ban_list.head, node, ban, ConfigItem_ban) {
		if (ban->flags & BAN_USERHOST) {
			if (!match(ban->mask, userhost)) {
				return conf_do_kline(ban, cptr, cptr);
			}
		}
		else if (ban->flags & BAN_IP) {
			if (!match(ban->mask, cptr->hostip)) {
				conf_do_zap(ban, cptr->hostip, cptr->localClient->fd);
				return exit_client(cptr, cptr, &me, "Zapped");
			}
		}
		else if (ban->flags & BAN_GCOS) {
			if (!match(ban->mask, cptr->info)) {
				return conf_do_gline(ban, cptr, cptr);
			}
		}
	}
	return 0;
}

static void check_list_pings(dlink_list *list)
{
	dlink_node *node, *next = NULL;
	aClient *cptr;
	int ping;

	DLINK_FOREACH_SAFE_DATA(list->head, node, next, cptr, aClient) {
		if (DeadSocket(cptr)) {
			exit_client(cptr, cptr, &me, SendQEx(cptr) ? "SendQ Exceeded" : "Dead Socket");
			continue;
		}
		if (Internal.do_ban_check && IsPerson(cptr)) {
			if (check_local_bans(cptr)) {
				continue;
			}
		}

		ping = (IsRegistered(cptr)) ? cptr->localClient->class->ping_time : CONNECTTIMEOUT;
		if (ping < (timeofday - cptr->lasttime)) {
			if ((PingSent(cptr) && (timeofday - cptr->lasttime) >= (ping * 2)) ||
			  (!IsRegistered(cptr) && (timeofday - cptr->since) >= ping)) {
				char timeout_str[32];

				if (IsServer(cptr) || IsConnecting(cptr) || IsHandshake(cptr)) {
					char *path = get_client_name(cptr, HIDE_IP);

					sendto_gnotice("No response from %s, closing link.", path);
					sendto_serv_msg_butone(cptr, &me, NULL, &CMD_GNOTICE,
						":No response from %s, closing link.", path);
				}

				ircsprintf(timeout_str, "Ping timeout: %d seconds",
					timeofday - cptr->lasttime);
				exit_client(cptr, cptr, &me, timeout_str);
				continue;
			}
			else if (!PingSent(cptr) && !IsConnecting(cptr) && !IsHandshake(cptr)
			  && (!IsServer(cptr) || (IsServer(cptr) && !SendingBurst(cptr)))) {
				SetPingSent(cptr);
				cptr->lasttime = timeofday - ping;
				sendto_one_client_nopostfix(cptr, NULL, &CMD_PING, ":%s", me.name);
			}
		}
	}
	Internal.do_ban_check = 0;
}

static void check_unknown_pings(dlink_list *list)
{
	dlink_node *node, *next = NULL;
	aClient *cptr;

	DLINK_FOREACH_SAFE_DATA(list->head, node, next, cptr, aClient) {
		if (cptr->firsttime ? ((timeofday - cptr->firsttime) > 30) : 0) {
			exit_client(cptr, cptr, &me, "Connection timed out");
		}
	}
}

static void check_pings()
{
	check_list_pings(&lclient_list);
	check_list_pings(&lserver_list);
	check_unknown_pings(&lunknown_list);
	Debug((DEBUG_DEBUG, "Next check_pings(): %s", myctime(timeofday + 9)));
}

void init_clients()
{
	client_heap = BlockHeapCreate(sizeof(aClient), CLIENTS_PREALLOCATE);
	localclient_heap = BlockHeapCreate(sizeof(LocalClient), LOCALCLIENTS_PREALLOCATE);

	user_heap = BlockHeapCreate(sizeof(anUser), USERS_PREALLOCATE);
	localuser_heap = BlockHeapCreate(sizeof(LocalUser), LOCALUSERS_PREALLOCATE);

	add_event("check_pings", check_pings, NULL, 9, 1);
}

aClient *next_client(aClient *next, char *ch)
{
	aClient *tmp = next;

	next = find_client(ch, tmp);
	if (tmp != NULL && tmp->prev == next) {
		return NULL;
	}
	if (next != tmp) {
		return next;
	}
	for (; next != NULL; next = next->next) {
		if (!match(ch, next->name)) {
			break;
		}
	}
	return next;
}

aClient *next_client_double(aClient *next, char *ch)
{
	aClient *tmp = next;

	next = find_client(ch, tmp);
	if (tmp != NULL && (tmp->prev == next)) {
		return NULL;
	}
	if (next != tmp) {
		return next;
	}
	for (; next != NULL; next = next->next) {
		if (!match(ch, next->name) || !match(next->name, ch)) {
			break;
		}
	}
	return next;
}

int use_or_deliver(aClient *cptr, aClient *sptr, msg_ptr *msg, char *format, int server, int parc,
  char *parv[])
{
	aClient *acptr;
	int wilds, localc = MyClient(sptr);
	char idbuf[16];

	if (parc <= server || BadPtr(parv[server]) || (localc && !match(me.name, parv[server]))) {
		return HUNTED_ISME;
	}
	if ((acptr = localc ? find_client(parv[server], NULL) : find_target(parv[server])) != NULL) {
		if ((!IsMe(acptr) && !IsRegistered(acptr)) || (acptr->from == sptr->from && !MyConnect(acptr))) {
			acptr = NULL;
		}
	}
	if (acptr == NULL) {
		collapse(parv[server]);

		if (!(wilds = has_wilds(parv[server]))) {
			if ((acptr = find_server(parv[server], NULL)) == NULL) {
				send_me_numeric(sptr, ERR_NOSUCHSERVER, parv[server]);
				return HUNTED_NOSUCH;
			}
		}
		else {
			for (acptr = client; (acptr = next_client(acptr, parv[server])) != NULL; acptr = acptr->next) {
				if (acptr->from == sptr->from && !MyConnect(acptr)) {
					continue;
				}
				if (IsRegistered(acptr) && acptr != cptr) {
					break;
				}
			}
		}
	}
	if (acptr != NULL) {
		if (IsMe(acptr) || MyClient(acptr)) {
			return HUNTED_ISME;
		}
		if (HasSUID(acptr) && CapID(acptr->from)) {
			idbuf[0] = '!';
			strcpy(&idbuf[1], acptr->id.string);
			parv[server] = idbuf;
		}
		else {
			parv[server] = acptr->name;
		}
		sendto_one_client_nopostfix(acptr, sptr, msg, format, parv[1], parv[2],
			parv[3], parv[4], parv[5], parv[6], parv[7], parv[8]);
		return HUNTED_PASS;
	}

	send_me_numeric(sptr, ERR_NOSUCHSERVER, parv[server]);
	return HUNTED_NOSUCH;
}

char *canonize(char *buffer)
{
	static char cbuf[BUFSIZE];
	char *s, *t, *cp = cbuf, *p = NULL, *p2;
	int l = 0;

	*cp = '\0';
	for (s = strtoken(&p, buffer, ","); s != NULL; s = strtoken(&p, NULL, ",")) {
		if (l > 0) {
			for (p2 = NULL, t = strtoken(&p2, cbuf, ","); t != NULL; t = strtoken(&p2, NULL, ",")) {
				if (!mycmp(s, t)) {
					break;
				}
				else if (p2 != NULL) {
					p2[-1] = ',';
				}
			}
		}
		else {
			t = NULL;
		}
		if (t == NULL) {
			if (l > 0) {
				*(cp - 1) = ',';
			}
			else {
				l = 1;
			}
			strcpy(cp, s);
			if (p != NULL) {
				cp += (p - s);
			}
		}
		else if (p2 != NULL) {
			p2[-1] = ',';
		}
	}
	return cbuf;
}

int register_user(aClient *cptr, aClient *sptr, char *nick, char *username)
{
	static char ubuf[16];
	char *parv[3], ipaddr[HOSTIPLEN + 1], tmp[512], *userhost, *p;
	aClient *nsptr;
	anUser *user = sptr->user;
	int bad_dns = 0;
	ConfigItem_ban *ban;
	HookData hdata;

	ASSERT(sptr != NULL);
	ASSERT(sptr->user != NULL);

	parv[0] = sptr->name;
	parv[1] = parv[2] = NULL;

	inetntop(&sptr->ip, ipaddr, HOSTIPLEN + 1);
	strncpyzt(sptr->hostip, ipaddr, HOSTIPLEN + 1);

	hdata.cptr = cptr;
	hdata.sptr = sptr;

	if (MyConnect(sptr)) {
		int auth;
		char *hostptr;

		ASSERT(sptr->localClient != NULL);
		ASSERT(sptr->localUser != NULL);

		sptr->localUser->last = timeofday;

		auth = check_client(sptr);
		hostptr = (*sptr->host != '\0') ? sptr->host : get_client_name(sptr, FALSE);

		if (auth != CLIENTAUTH_MATCHED) {
			ircstp->is_ref++;
		}
		switch (auth) {
			case CLIENTAUTH_NOMATCH:
				sendto_realops_lev(REJ_LEV, CLIENTAUTH_ERRMSG, hostptr, "");
				ircdlog(LOG_CLIENT, CLIENTAUTH_ERRMSG, hostptr, "");
				return exit_client(cptr, sptr, &me, "You are not authorized to connect to "
					"this server");
				break;
			case CLIENTAUTH_ILINEFULL:
				sendto_realops_lev(REJ_LEV, CLIENTAUTH_ERRMSG, hostptr, " (I line is full)");
				ircdlog(LOG_CLIENT, CLIENTAUTH_ERRMSG, hostptr, " (I line is full)");
				return exit_client(cptr, sptr, &me, "Your connection class is full");
				break;
			case CLIENTAUTH_INVALIDPW:
				sendto_realops_lev(REJ_LEV, CLIENTAUTH_ERRMSG, hostptr, " (invalid/no password)");
				ircdlog(LOG_CLIENT, CLIENTAUTH_ERRMSG, hostptr, " (invalid/no password)");
				return exit_client(cptr, sptr, &me, "Invalid/No password");
				break;
			case CLIENTAUTH_TOOMANYIPS:
				sendto_realops_lev(REJ_LEV, CLIENTAUTH_ERRMSG, hostptr, " (too many connections)");
				ircdlog(LOG_CLIENT, CLIENTAUTH_ERRMSG, hostptr, " (too many connections)");
				return exit_client(cptr, sptr, &me, "Too many connections from your IP address");
				break;
		}
		for (p = sptr->host; *p != '\0'; p++) {
			if (!isalnum(*p)) {
				if (*p != '-' && *p != '.') {
					bad_dns = 1;
				}
			}
		}
		if (bad_dns) {
			send_me_noticeNA(sptr, ":*** You have a bad character in your hostname, using IP "
				"address instead.");
			strcpy(sptr->host, sptr->hostip);
			strcpy(sptr->localClient->sockhost, sptr->hostip);
		}

		if ((sptr->localClient->class->flags & CLASS_IGNOREIDENTD) || (!GotID(sptr)
		  && (sptr->localClient->class->flags & CLASS_NOTILDE)
		  && !(sptr->localClient->class->flags & CLASS_REQUIREIDENTD))) {
			if (sptr->username != username) {
				strncpyzt(sptr->username, username, USERLEN + 1);
			}
		}
		else if (!GotID(sptr)) {
			if (sptr->localClient->class->flags & CLASS_REQUIREIDENTD) {
				send_me_noticeNA(sptr, ":*** Your connection class requires your host to "
					"have identd installed in order to connect.");
				return exit_client(cptr, sptr, &me, "Your connection class requires identd");
			}
			else {
				char tmp2[USERLEN + 1];
				strncpyzt(tmp2, username, USERLEN + 1);
				*sptr->username = '~';
				strncpy(&sptr->username[1], tmp2, USERLEN);
				sptr->username[USERLEN] = '\0';
			}
		}

		if (sptr->localClient->fd >= (ServerInfo->max_clients + MAX_BUFFER)
		  || (sptr->localClient->fd >= (ServerInfo->max_clients - 5))) {
			sendto_realops_lev(REJ_LEV, "Too many clients, rejecting %s[%s]", nick,
				sptr->localClient->sockhost);
			ircstp->is_ref++;
			return exit_client(cptr, sptr, &me, "Sorry, server is full - try later");
		}

		userhost = make_user_host(sptr->username, sptr->host);
		if ((ban = find_ban(BAN_USERHOST, userhost)) != NULL) {
			return conf_do_kline(ban, cptr, sptr);
		}

		{
			char *str;
			unsigned char c;
			int special = 0;

			if (strchr(sptr->username, '@') || strchr(username, '@')) {
				sendto_realops_lev(REJ_LEV, "Illegal \"@\" in username: %s (%s)",
					get_client_name(sptr, FALSE), username);
				ircstp->is_ref++;
				ircsprintf(tmp, "Invalid username [%s] - '@' is not allowed!", username);
				return exit_client(cptr, sptr, sptr, tmp);
			}

			str = (*sptr->username == '~') ? (sptr->username + 1) : sptr->username;
			while (*str != '\0') {
				c = *str++;
				if ((!isalnum(c) && !strchr(" -_.", c)) || (c > 127) || (c < 32)) {
					special++;
				}
			}
			if (special) {
				sendto_realops_lev(REJ_LEV, "Invalid username: %s (%s@%s)", nick,
					sptr->username, sptr->host);
				ircstp->is_ref++;
				ircsprintf(tmp, "Invalid username [%s]", sptr->username);
				return exit_client(cptr, sptr, &me, tmp);
			}

			if (irccmp(sptr->username, username)) {
				str = (*username == '~') ? (username + 1) : username;
				while (*str != '\0') {
					c = *str++;
					if ((!isalnum(c) && !strchr(" -_.", c)) || (c > 127)) {
						special++;
					}
				}
				if (special) {
					sendto_realops_lev(REJ_LEV, "Invalid username: %s (%s@%s)",
						nick, username, sptr->host);
					ircstp->is_ref++;
					ircsprintf(tmp, "Invalid username [%s]", username);
					return exit_client(cptr, sptr, &me, tmp);
				}
			}
		}

		if (*(sptr->username + 1) == '\0' && !isalpha(*sptr->username)) {
			sendto_realops_lev(REJ_LEV, "Invalid username: %s (%s@%s)", nick,
				sptr->username, sptr->host);
			ircstp->is_ref++;
			ircsprintf(tmp, "Invalid username [%s]", sptr->username);
			return exit_client(cptr, sptr, &me, tmp);
		}

		if (hook_run_until(h_post_access, &hdata, FLUSH_BUFFER) == FLUSH_BUFFER) {
			return FLUSH_BUFFER;
		}

		ircdlog(LOG_CLIENT, "client connecting on port %d: %s", sptr->localClient->listener->port,
			get_client_name(sptr, FALSE));
		sendto_realops_lev(CCONN_LEV, "Client connecting: %s (%s@%s) {%s}", nick,
			sptr->username, sptr->host, sptr->localClient->class->name);

		if (++Count.local > Count.max_loc) {
			Count.max_loc = Count.local;
		}
		if (timeofday - Count.day > 86400) {
			Count.today = 0;
			Count.day = timeofday;
		}
		if (timeofday - Count.week > 604800) {
			Count.weekly = 0;
			Count.week = timeofday;
		}
		if (timeofday - Count.month > 2592000) {
			Count.monthly = 0;
			Count.month = timeofday;
		}
		if (timeofday - Count.year > 31536000) {
			Count.yearly = 0;
			Count.year = timeofday;
		}
		Count.today++;
		Count.weekly++;
		Count.monthly++;
		Count.yearly++;
	}
	else {
		strncpyzt(sptr->username, username, USERLEN + 1);
	}

	SetClient(sptr);
	strncpyzt(sptr->user->maskedhost, maskme(sptr), HOSTLEN + 1);

	if (++Count.total > Count.max_tot) {
		Count.max_tot = Count.total;
	}
	if (HasMode(sptr, UMODE_INVISIBLE)) {
		Count.invisi++;
	}

	if (MyConnect(sptr)) {
		dlink_del_nofree(&lunknown_list, NULL, &sptr->localClient->self);
		dlink_add_node(&lclient_list, &sptr->localClient->self, sptr);
		Count.unknown--;

		if (assign_local_userid(sptr)) {
			SetSUID(sptr);
		}

		Count.cli_restart++;

#ifdef INCREASE_SOCK_BUFS
		increase_sock_bufs(sptr->localClient->fd, 0);
#endif

		send_me_numeric(sptr, RPL_WELCOME, NetworkConfig.name, sptr->name, sptr->username,
			sptr->host);
		send_me_numeric(sptr, RPL_YOURHOST, me.name, sptr->localClient->listener->ipaddr,
			sptr->localClient->listener->port, ircd_version);
#ifdef IRCII_KLUDGE
		send_me_notice(sptr, ":*** Your host is %s, running version %s",
			get_client_name(&me, SHOW_IP), ircd_version);
#endif
		send_me_numeric(sptr, RPL_CREATED, creation);

		if (sptr->localClient->listener->flags & LISTEN_CRJAVA) {
			send_me_numeric_buf(sptr, "%s CR1.8.03-%s %s %s", RPL_MYINFO, me.name,
				ircd_version, usermode_str, chanmode_str);
		}
		else {
			send_me_numeric(sptr, RPL_MYINFO, me.name, ircd_version, usermode_str,
				chanmode_str);
		}

		send_me_numeric(sptr, RPL_ISUPPORT, isupport1);
		send_me_numeric(sptr, RPL_ISUPPORT, isupport2);
		send_me_numeric(sptr, RPL_ISUPPORT, isupport3);

#ifdef USE_OPENSSL
		if (IsSecure(sptr) && (sptr->localClient->ssl != NULL)) {
			send_me_notice(sptr, ":*** You are connected via %s.",
				ssl_get_cipher_info(sptr->localClient->ssl));
		}
#endif

		send_lusers(sptr, sptr, 1, parv);

		if (GeneralConfig.short_motd) {
			send_me_noticeNA(sptr, ":*** Please read the full motd if you haven't read it.");
			send_me_numeric(sptr, RPL_MOTDSTART, me.name);
			send_me_numeric(sptr, RPL_MOTD, "*** This is the short motd ***");
			send_me_numericNA(sptr, RPL_ENDOFMOTD);
		}
		else {
			send_motd(sptr);
		}

		hook_run(h_post_motd, &hdata);
	}
	else if (IsServer(cptr)) {
		aClient *acptr;

		if ((acptr = find_server(user->server, NULL)) == NULL) {
			sendto_one_client(sptr, &me, &CMD_KILL, ":%s GHOST (no server %s on this net)",
				user->server);
			sendto_realops_lev(DEBUG_LEV, "No server %s for user %s[%s@%s] from %s",
				user->server, sptr->name, sptr->username, sptr->host,
				sptr->from->name);
			SetKilled(sptr);
			return exit_client(sptr, sptr, &me, "Ghosted Client");
		}
		else if (acptr != NULL && (acptr->from != sptr->from)) {
			sendto_one_client(sptr, &me, &CMD_KILL, ":%s (%s != %s[%s] USER from wrong direction)",
				me.name, user->server, acptr->from->name, acptr->from->localClient->sockhost);
			sendto_realops_lev(DEBUG_LEV, "Bad User [%s] :%s USER %s@%s %s, != %s[%s]",
				cptr->name, nick, sptr->username, sptr->host, user->server,
				acptr->name, acptr->from->name);
			SetKilled(sptr);
			return exit_client(sptr, sptr, &me, "USER server wrong direction");
		}
		if (!IsULine(sptr) && !SendingBurst(cptr)) {
			sendto_realops_lev(GCCONN_LEV, "Client connecting at %s: %s (%s@%s)",
				sptr->user->server, sptr->name, sptr->username, sptr->host);
		}
	}

	send_umode(NULL, sptr, 0, SEND_UMODES, ubuf);
	if (*ubuf == '\0') {
		ubuf[0] = '+';
		ubuf[1] = '\0';
	}

	hash_check_watch(sptr, RPL_LOGON);
	send_nick_to_serv_butone(cptr, sptr, ubuf);

	if (MyClient(sptr)) {
		if ((nsptr = find_person(NICKSERV, NULL)) != NULL && (*sptr->localClient->passwd != '\0')) {
			sendto_one_client_onserver(nsptr, sptr, &CMD_PRIVMSG, ":IDENTIFY %s",
				sptr->localClient->passwd);
		}

		memset(sptr->localClient->passwd, '\0', PASSWDLEN + 1);

		if (ubuf[1] != '\0') {
			send_umode(cptr, sptr, 0, ALL_UMODES, ubuf);
		}

		hook_run(h_post_register_user, &hdata);
	}
	return 0;
}

int do_user(char *nick, aClient *cptr, aClient *sptr, char *username, char *host,
  char *server, unsigned long serviceid, unsigned int ip, char *realname)
{
	anUser *user = make_user(sptr);

	if (!MyConnect(sptr)) {
		user->server = find_or_add(server);
		strncpyzt(sptr->host, host, HOSTLEN + 1);
	}
	else {
		if (!IsUnknown(sptr)) {
			send_me_numericNA(sptr, ERR_ALREADYREGISTRED);
			return 0;
		}

		make_localuser(sptr);
		sptr->localUser->oflags = 0;

		user->server = me.name;
		AddMode(sptr, GeneralConfig.modes_on_connect);
	}

	strncpyzt(sptr->info, realname, sizeof(sptr->info));
	sptr->user->servicestamp = serviceid;

	if (!MyConnect(sptr)) {
		sptr->ip.s_addr = ntohl(ip);
#ifdef USE_THROTTLE
		if (ip) {
			throttle_check((char *)inetntoa((const char *)&sptr->ip), -1, sptr->tsinfo);
		}
#endif
	}

	if (*sptr->name != '\0') {
		return register_user(cptr, sptr, sptr->name, username);
	}
	else {
		strncpyzt(sptr->username, username, USERLEN + 1);
	}
	return 0;
}

int is_silenced(aClient *sptr, aClient *acptr)
{
	SLink *lp;
	char *sender;

	if (acptr->user == NULL || acptr->user->silence == NULL || sptr->user == NULL) {
		return 0;
	}

	sender = make_nick_user_host(sptr->name, sptr->username, sptr->host);
	for (lp = acptr->user->silence; lp != NULL; lp = lp->next) {
		if (!match(lp->value.cp, sender)) {
			if (!MyConnect(sptr)) {
				sendto_one_client(sptr, acptr, &CMD_SILENCE, "%s", lp->value.cp);
				lp->flags = 1;
			}
			return 1;
		}
	}
	return 0;
}

int del_silence(aClient *sptr, char *mask)
{
	SLink **lp, *tmp;
	for (lp = &(sptr->user->silence); *lp != NULL; lp = &((*lp)->next)) {
		if (!mycmp(mask, (*lp)->value.cp)) {
			tmp = *lp;
			*lp = tmp->next;
			MyFree(tmp->value.cp);
			free_slink(tmp);
			return 0;
		}
	}
	return 1;
}

int add_silence(aClient *sptr, char *mask)
{
	SLink *lp;
	int cnt = 0, len = 0;

	for (lp = sptr->user->silence; lp !=  NULL; lp = lp->next) {
		len += strlen(lp->value.cp);
		if (MyClient(sptr)) {
			if (len > MAXSILELENGTH || ++cnt >= MAXSILES) {
				send_me_numeric(sptr, ERR_SILELISTFULL, mask);
				return -1;
			}
			if (!match(lp->value.cp, mask)) {
				return -1;
			}
		}
		else if (!mycmp(lp->value.cp, mask)) {
			return -1;
		}
	}

	lp = make_slink();
	lp->next = sptr->user->silence;
	lp->value.cp = (char *)MyMalloc(strlen(mask) + 1);
	strcpy(lp->value.cp, mask);
	sptr->user->silence = lp;
	return 0;
}
